namespace eval VELaSSCo {
    variable _DEBUG_MODE          0    ;# set to 1 to show the ShowHideLogWindow icon
    variable _dir
    variable _images
    variable _tunnel_used 0
    variable _sessionID   ""
    variable _modelID     ""
    variable _modelInfo
    variable _status_wid  ""
    variable _login_icon_wid     ""
    variable _icon_do_connection connect_creating.png
    variable _icon_connected     connect_established.png
    # due to a change in Postprocess done by Enrique in 12.1.10d this can not be used anymore:
    # variable _dummy_mesh_name    dummyMesh_*back*
    variable _dummy_mesh_name    dummyMesh_back
    # variable _dummy_result_value 1e-38
    variable _dummy_result_value 0
    variable _cache
    variable _selected
    variable _busy
    variable _d2c
    variable _preferences
    variable _pending            {}
    variable _ogl
    # maximum vervosity
    variable _verbose_level      9
    # verbose levels:
    # 0 - Nothing
    # 1 - show queries issued with short response

    # utils, Ini & End
    proc GetFullImageName { filename} {
	variable _dir
	return [ file join $_dir images $filename]
    }
    proc GetImage { filename} {
	variable _dir
	variable _images
	if { ![ info exist _images($filename)]} {
	    set _images($filename) [ image create photo -file \
					 [ file join $_dir images $filename]]
	} 
	return $_images($filename)
    }

    proc Ini { dir} {
	variable _dir

	set _dir $dir

	set VersionRequired "12.0.0"
	set comp -1
	catch { 
	    set comp [ GiDVersionCmp $VersionRequired]
	}

	# doing package require ...
	if { [ string first $_dir $::auto_path] == -1} {
	    lappend ::auto_path ${_dir}
	}

	set err_txt ""
	set err [ catch {
	    set ret [ package require gid_velassco]
	    # W "package require gid_velassco = $ret"
	} err_txt ]
	if { $err} {
	    # problems loading the package ( i.e. library)
	    W $err_txt
	    if { $::tcl_platform(platform) == "windows"} {
		set suggestion "\nThe package has been compiled with\n"
		append suggestion "Microsoft Visual C++ 2012\n"
		append suggestion "May be the solution is to install the\n"
		append suggestion "Microsoft Visual C++ 2012 Redistributable (x64)\n"
		append suggestion "which can be found at [ file join $_dir bin vcredist_x64.exe]"
		W $suggestion
		WarnWin $suggestion
	    } else {
		W "Dependencies:"
		set lib_name [ file join $_dir gid_velassco [ Get_Library_gid_velassco_name gid_velassco]]
		W [ exec ldd $lib_name]
		set suggestion "\nMay be the solution is to install\n"
		append suggestion "boost >= 1.58\n"
		append suggestion "thrift >= 0.92\n"
		append suggestion "openssl >= 1.0.1p\n"
		append suggestion "libevent >= 2.0"
		W $suggestion
	    }
	}

	# ReadPreferences
	return
    }

    proc End { args } {
	variable _DEBUG_MODE
	DoLogout
	# the Log window:
	if { $_DEBUG_MODE} {
	    if { [info exists ::RunProcInfo] && ( $::RunProcInfo == "")} {
		ShowHideLogWindow
	    }
	}
    }

    # proc GetPreferencesFileName {} {
    # 	set gid_defaults [GiveGidDefaultsFile]; # gid.ini / .gidDefaults
    # 	set pref_dir [ file dirname $gid_defaults]
    # 	set velassco_prefs_file [ file join $pref_dir .gid_velassco]
    # 	return $velassco_prefs_file
    # }
    # 
    # proc ReadPreferences { } {
    # 	variable _preferences
    # 	set velassco_prefs_file [ GetPreferencesFileName]
    # 	if { [ file exists $velassco_prefs_file]} {
    # 	    # read file
    # 	    set fi [ open $velassco_prefs_file]
    # 	    while { ![ eof $fi] } {
    # 		set line [ string trimleft [ gets $fi]]
    # 		if { ( [ string length $line] == 0) || ( [ string index $line 0] == "#") || ( [ llength $line] < 2)} {
    # 		    continue
    # 		}
    # 		set idx [ lindex $line 0]
    # 		set val [ lindex $line 1]
    # 		set _preferences($idx) $val
    # 	    }
    # 	    close $fi
    # 	} else {
    # 	    # W "File '$velassco_prefs_file' not found"
    # 	}
    # }

    proc ::VELaSSCo::GetPreferencesFromVELaSSCo {} {
	variable _preferences
	variable _sessionID

	if { ![ CheckLogin]} { return}

	# get all configuration parameters from velassco:
	set err_txt "Ok"
	set cmd "gid_velassco GetConfiguration $_sessionID"
	set lst_param_value ""
	set err [ catch {
	    set lst_param_value [ {*}$cmd]
	} err_txt]
	if { !$err} {
	    set err_txt "Ok"
	}
	Verbose 1 "$cmd = ($err_txt) $lst_param_value"
	# W -$lst_param_value-
	if { !$err} {
	} else {
	    W "Error doing $cmd = $err_txt"
	}	
	# lst_param_value is something like:
	# CompressionType
	# ZLIB
	# CompressionLevel
	# 1
	# CompressionThreshold
	# 1048576
	# CompressionEnabled
	# 0
	# CompressionList
	# Raw Zlib Lzo
	set lst [ split $lst_param_value \n]
	foreach [ list idx val] $lst {
	    # skip empty indices / values
	    if { [ string length $idx] >= 1} {
		set _preferences($idx) $val
		# W "set _preferences($idx) $val"
	    }
	}

	# node volution number of steps preferences
	if { ![ info exists _preferences(NumberOfStepsForNode)]} {
	    set _preferences(NumberOfStepsForNode) 5
	}
    }

    proc SavePreferences { } {
	variable _preferences

	set velassco_prefs_file [ GetPreferencesFileName]
	set fo [ open ${velassco_prefs_file}.new w]
	foreach idx [ array names _preferences] {
	    puts $fo "[ list $idx] [ list $_preferences($idx)]"
	}
	close $fo
	file rename -force ${velassco_prefs_file} ${velassco_prefs_file}.bak
	file rename -force ${velassco_prefs_file}.new ${velassco_prefs_file}
    }

    proc ApplyPreferences {} {
	variable _preferences
	variable _sessionID
	variable _cache

	if { ![ CheckLogin]} { return}
	
	foreach idx [ list CompressionEnabled CompressionType CompressionLevel CompressionThreshold] {
	    set cmd "gid_velassco SetConfiguration $_sessionID $idx $_preferences($idx)"
	    set ret ""
	    set err_txt "Ok"
	    set err [ catch {
		set ret [ {*}$cmd]
	    } err_txt]
	    if { !$err} {
		set err_txt "Ok"
	    }
	    Verbose 1 "$cmd = ($err_txt) $ret"
	}
	# Nothing to do: node volution number of steps preferences
	# set _preferences(NumberOfStepsForNode) 5
	if { ( $::VELaSSCo::_preferences(newNumberOfStepsForNode) == "all") ||
	     ( [ string is integer $::VELaSSCo::_preferences(newNumberOfStepsForNode)] &&
	       ( $::VELaSSCo::_preferences(newNumberOfStepsForNode) > 0))} {
	    set ::VELaSSCo::_preferences(NumberOfStepsForNode) $::VELaSSCo::_preferences(newNumberOfStepsForNode)
	    Verbose 1 "Number of steps for the evolution graph of a vertex' result set to '$::VELaSSCo::_preferences(newNumberOfStepsForNode)'"
	} else {
	    ErrorWin "Wrong number of steps for node evolution graph:\n$::VELaSSCo::_preferences(newNumberOfStepsForNode)\nShould be > 0 or 'all'"
	}
    }

    proc PreferencesWindow { { w .gid.wVelasscoPreferences}} {
	variable _preferences

	if { ![ CheckLogin]} { return}

	GetPreferencesFromVELaSSCo
	
	# create window
	if { [ winfo exists $w]} {
	    destroy $w
	}
	toplevel $w
	wm title $w [_ "VELaSSCo preferences ..."]
	CreateDecorationWindow $w  $w.top $w.bot\
	    [ list [_ "Apply"] [_ "Close"]] \
	    [ list [ list VELaSSCo::ApplyPreferences ] [ list destroy $w]] \
	    [ list <Return> <Escape>]
	ttk::frame $w.top.header
	ttk::label $w.top.header.l1 -image [ VELaSSCo::GetImage VelasscoBig67x48.png]
	ttk::label $w.top.header.l2 -text [_ "VELaSSCo preferences"]
	grid $w.top.header.l1 $w.top.header.l2 -padx 4 -pady 4 -sticky ew
	
	ttk::frame $w.top.preferences

	# Brute force:
	# set w_idx 1
	# foreach idx [ array names _preferences] {
	#     ttk::label $w.top.preferences.l$w_idx -text $idx
	#     ttk::entry $w.top.preferences.e$w_idx -textvariable ::VELaSSCo::_preferences($idx)
	# 
	#     grid $w.top.preferences.l$w_idx $w.top.preferences.e$w_idx -sticky e -padx 2
	#     grid configure $w.top.preferences.e$w_idx -sticky ew
	#     incr w_idx
	# }

	# Nice:
	# CompressionType
	# ZLIB
	# CompressionLevel
	# 1
	# CompressionThreshold
	# 1048576
	# CompressionEnabled
	# 0
	# CompressionList
	# Raw Zlib

	set lst_w_compression {}
	ttk::labelframe $w.top.preferences.comm -text [_ "Connection options "]
	ttk::checkbutton $w.top.preferences.comm.cb_compress \
	    -text [_ "Compress incomming results."] \
	    -variable ::VELaSSCo::_preferences(CompressionEnabled)

	ttk::label $w.top.preferences.comm.l_comp_type -text [_ "Compression type:"]
	set lst_labels {}
	foreach val $_preferences(CompressionList) {
	    lappend lst_labels [ string totitle $val]
	}
	set _preferences(CompressionType) [ string totitle [ string trimright $_preferences(CompressionType)]]
	proc ::VELaSSCo::EnableDisableCompLevelOptions { w } {
	    if { $::VELaSSCo::_preferences(CompressionEnabled)} {
		set state normal
		set lst_disabled_types [ list raw lzo]
		set comp_type [ string tolower $::VELaSSCo::_preferences(CompressionType)]
		if { [ lsearch $lst_disabled_types $comp_type] != -1} {
		    set state disabled
		}
		$w.top.preferences.comm.l_comp_level configure -state $state
		$w.top.preferences.comm.s_comp_level configure -state $state
	    }
	}
	TTKMenubutton $w.top.preferences.comm.mb_comp_type \
	    -labels $lst_labels \
            -textvariable ::VELaSSCo::_preferences(CompressionType) \
	    -modifycmd "::VELaSSCo::EnableDisableCompLevelOptions $w"
	lappend lst_w_compression $w.top.preferences.comm.l_comp_type $w.top.preferences.comm.mb_comp_type

	ttk::label $w.top.preferences.comm.l_comp_level -text [_ "Compression level (Zlib):"]
	gidscale $w.top.preferences.comm.s_comp_level \
	    -orient horizontal -from 1 -to 9 -resolution 1 \
	    -showbuttons 1 -symbolbuttons "- +" \
	    -showvalue 1 \
	    -variable ::VELaSSCo::_preferences(CompressionLevel)
	lappend lst_w_compression $w.top.preferences.comm.l_comp_level $w.top.preferences.comm.s_comp_level

	label $w.top.preferences.comm.l_comp_min_mem -text [_ "Minimum size to compress:"]
	# ttk::entry $w.top.preferences.comm.e_comp_min_mem \
	#    -textvariable ::VELaSSCo::_preferences(CompressionThreshold) -width 8
	set lst_labels [ list 10KBytes 20KBytes 50KBytes 100KBytes 200KBytes 500KBytes 1MBytes 2MBytes 5MBytes 10MBytes]
	set lst_values [ list 10240 20480 51200 102400 204800 512000 1048576 2097152 5242880 10485760]
	TTKMenubutton $w.top.preferences.comm.e_comp_min_mem \
	    -labels $lst_labels \
	    -values $lst_values \
	    -textvariable ::VELaSSCo::_preferences(CompressionThreshold)
	lappend lst_w_compression $w.top.preferences.comm.l_comp_min_mem $w.top.preferences.comm.e_comp_min_mem

	grid $w.top.preferences.comm.cb_compress -columnspan 2 -sticky w -padx 2
	grid $w.top.preferences.comm.l_comp_type $w.top.preferences.comm.mb_comp_type -sticky e -padx 2
	grid configure $w.top.preferences.comm.mb_comp_type -sticky w
	grid $w.top.preferences.comm.l_comp_level $w.top.preferences.comm.s_comp_level -sticky e -padx 2
	grid configure $w.top.preferences.comm.s_comp_level -sticky ew
	grid $w.top.preferences.comm.l_comp_min_mem $w.top.preferences.comm.e_comp_min_mem -sticky e -padx 2
	grid configure $w.top.preferences.comm.e_comp_min_mem -sticky ew
	grid columnconfigure $w.top.preferences.comm 1 -weight 1
	grid rowconfigure $w.top.preferences.comm 1 -weight 1


	ttk::labelframe $w.top.preferences.other -text [_ "Graphs "]
	ttk::label $w.top.preferences.other.l_evolution_steps_number -text [_ "Number of steps for evolution graph:"]
	# ttk::entry $w.top.preferences.e_evolution_steps_number -textvariable ::VELaSSCo::_preferences(newNumberOfStepsForNode)
	TTKMenubutton $w.top.preferences.other.e_evolution_steps_number \
	    -values [ list 5 15 25 all] \
	    -width 5 \
	    -textvariable ::VELaSSCo::_preferences(newNumberOfStepsForNode)
	set ::VELaSSCo::_preferences(newNumberOfStepsForNode) $::VELaSSCo::_preferences(NumberOfStepsForNode)

	grid $w.top.preferences.other.l_evolution_steps_number $w.top.preferences.other.e_evolution_steps_number -sticky e -padx 2
	grid configure $w.top.preferences.other.e_evolution_steps_number -sticky w

	grid $w.top.preferences.comm -sticky news
	grid $w.top.preferences.other -sticky news

	grid columnconfigure $w.top.preferences 0 -weight 1
	# grid rowconfigure $w.top 1 -weight 1

	grid $w.top.header -sticky ew
	grid $w.top.preferences -sticky ewns
	grid columnconfigure $w.top 0 -weight 1
	grid rowconfigure $w.top 1 -weight 1

	::VELaSSCo::EnableDisableCompLevelOptions $w
	VELaSSCo::EnableDisableWidgets $lst_w_compression ::VELaSSCo::_preferences(CompressionEnabled)
	$w.top.preferences.comm.cb_compress configure \
	    -command "VELaSSCo::EnableDisableWidgets \"$lst_w_compression\" ::VELaSSCo::_preferences(CompressionEnabled) ; ::VELaSSCo::EnableDisableCompLevelOptions $w "

    }

    proc WaitState { } {
	::GidUtils::WaitState .gid
    }
    
    proc EndWaitState { } {
	::GidUtils::EndWaitState .gid
    }

    proc Print { msg } {
	# W $msg
	GidUtils::SetWarnLine $msg
	# update to show the message to the user
	update
    }

    proc Verbose { level msg} {
	variable _verbose_level
	if { $level < $_verbose_level} {
	    Print $msg
	}
    }

    proc Debug { msg } {
	# W $msg
	GidUtils::SetWarnLine $msg
    }

    proc ShowChanges {} {
	variable _dir
	set readme [ file join $_dir README.txt]
	set readme_txt ""
	set err [ catch {
	    set fi [ open $readme]
	    set readme_txt [ read $fi]
	    close $fi
	} err_txt]
	if { $err} {
	    W $err_txt
	}
	W $readme_txt "ChangeLog"
    }

    # functionalities / Vqueries
    proc About { } {
	set package_version "N/A"
	catch {
	    set package_version [ package require gid_velassco]
	}
	# InfoWin "About VELaSSCo:\n\nUsing gid_velassco v. $package_version\nwith support for\n[ gid_velassco help]\n\nPlease visit http://www.velassco.eu"

	set w .gid.wVelasscoAbout
	if { [ winfo exists $w]} {
	    destroy $w
	}
	toplevel $w
	wm title $w [_ "About VELaSSCo..."]
	CreateDecorationWindow $w  $w.top $w.bot\
	    [ list [_ "Changes..."] [_ "Ok"]] \
	    [ list ::VELaSSCo::ShowChanges [ list destroy $w]] \
	    [ list "" <Return>]
	ttk::frame $w.top.header
	ttk::label $w.top.header.l1 -image [ VELaSSCo::GetImage VelasscoBig67x48.png]
	ttk::label $w.top.header.l2 -text [_ "About VELaSSCo"]
	grid $w.top.header.l1 $w.top.header.l2 -padx 4 -pady 4 -sticky ew

	ttk::frame $w.top.about
	ttk::label $w.top.about.l0 -text [_ "Using gid_velassco v. $package_version"]
	ttk::label $w.top.about.l1 -text [_ "with support for:"]
        ttk::scrollbar $w.top.about.scrolly -command "$w.top.about.t yview" -orient vertical
	text $w.top.about.t -wrap word -width 30 \
	    -yscrollcommand "$w.top.about.scrolly set" \
	    -border 0 -background [ $w cget -background]
	# -height 8
	VisitWebButton $w.top.about.b http://www.velassco.eu "Please visit www.velassco.eu"

	grid $w.top.about.l0 -sticky w -padx 4 -pady 0 -columnspan 2
	grid $w.top.about.l1 -sticky w -padx 4 -pady 0 -columnspan 2
	grid $w.top.about.t -sticky news -padx 4 -pady 0
	grid $w.top.about.scrolly -row 2 -column 1 -sticky ns
	grid $w.top.about.b -sticky ew -padx 4 -pady 0 -columnspan 2
	grid columnconfigure $w.top.about 0 -weight 1
	grid rowconfigure $w.top.about 2 -weight 1
	# grid rowconfigure $w.top.about 1 -weight 1

	grid $w.top.header -sticky ew
	grid $w.top.about -sticky ew -padx 4 -pady 4
	grid columnconfigure $w.top 0 -weight 1
	grid rowconfigure $w.top 1 -weight 1

	# get rid of gid_velassco ... ?parameters?
	regsub -all { \| } [ lrange [ gid_velassco help] 1 end-1] \n txt
	$w.top.about.t insert end $txt
	# $w.top.about.t insert end ""
	# $w.top.about.t see end
	$w.top.about.t configure -state disabled
    }

    proc OpenModel { { w .gid.wVelasscoOpenModel}} {
	variable _sessionID
	variable _modelID
	variable _modelInfo

	if { ![ CheckLogin]} { return}
	# Debug "VQuery OpenModel( $::VELaSSCo(SelectedModel))"
	if { "$::VELaSSCo(SelectedModel)" != ""} {
	    # should verify if there is a model already opened
	    if { "$_modelID" != ""} {
		CloseModel
	    }
	    set err_txt "Ok"
	    set err [ catch {
		set _modelID [ gid_velassco OpenModel $_sessionID $::VELaSSCo(SelectedModel) rw]
	    } err_txt]
	    if { !$err} {
		set err_txt "Ok"
	    }
	    Verbose 1 "gid_velassco OpenModel $_sessionID $::VELaSSCo(SelectedModel) rw = ($err_txt) $_modelID"
	    if { !$err} {
		wm title .gid $::VELaSSCo(SelectedModel)
		set _modelInfo(unique_name) $::VELaSSCo(SelectedModel)
	    } else {
		# clean any Postprocess information inside GiD:
		GiD_Process Mescape Files New Yes escape
	    }
	}
    }

    proc CloseModel {} {
	variable _sessionID
	variable _modelID
	variable _modelInfo
	variable _cache
	variable _selected
	variable _busy
	variable _pending
	variable _d2c
	if { ![ CheckLogin]} { return}
	if { "$_modelID" != ""} {
	    set err_txt "Ok"
	    set ret ""
	    set err [ catch {
		set ret [ gid_velassco CloseModel $_sessionID $_modelID]
	    } err_txt]
	    if { !$err} {
		set err_txt "Ok"
	    }
	    Verbose 1 "gid_velassco CloseModel $_sessionID $_modelID = ($err_txt) $ret"
	    set _modelID ""
	    OGLunregisterAndDeleteLists
	    GiD_Process Mescape Files New Yes escape
	    set back $_cache(Models,Names)
	    catch {
		array unset _modelInfo
		array unset _cache
		array unset _selected
		array unset _busy
		array unset _pending
		array unset _d2c
	    }
	    set _cache(Models,Names) $back
	    set ::VELaSSCo::_warning_gp_once 0
	}
    }

    proc DoSelectModel { w} {
	variable _sessionID
	variable _modelID
	
	if { $::VELaSSCo(SelectedModel) != ""} {
	    destroy $w
	    OpenModel
	    Verbose 1 "selected model = modelID = $_modelID"
	    # loading information about the model automatically as it's needed
	    GetModelInfo 1
	} else {
	    ErrorWin [_ "No model has been selected"]
	}
    }

    proc DoUpdateThumbImage { wThumb img} {
	if { [ winfo exists $wThumb]} {
	    $wThumb configure -text $img
	}
    }

    proc GetAndUpdateThumbnail { x y w tbl wThumb} {

	# set tablelist::x [expr {$x + [winfo x $w]}]
	# set tablelist::y [expr {$y + [winfo y $w]}]
	# set cell [$tbl nearestcell $tablelist::x $tablelist::y]
	# foreach {row col} [ split $cell ,] break    

	set setsel [$tbl curselection]

	set model_unique_name "VELaSSCo_Models:*:*"
	if { $setsel == {} } {
	    tk_dialogRAM $w.terr error [_ "Sorry, no model selected."] error 0 [_ "OK"]
	} else {
	    if { [ llength $setsel] == 1} {
		set row_values [ $tbl get $setsel]
		set model_name [ lindex $row_values 0]
		set model_path [ lindex $row_values 1]
		set model_loc [ lindex $row_values 2]
		# location can be of the form HBase:TableName or EDM
		if { [ string compare -nocase "EDM" $model_loc] == 0 } {
		    # it matches "EDM"
		    set model_unique_name $model_name
		} else {
		    # if location == HBASE:TableName then location has the table name
		    # otherwise don't use it, as we don't know
		    set location_lst [ split $model_loc :]
		    set location_lst_size [ llength $location_lst]
		    if { ( $location_lst_size >= 1) && [ string equal -nocase [ lindex $location_lst 0] "hbase"]} {
			if { $location_lst_size == 2} {
			    set model_unique_name "[ lindex $location_lst 1]:$model_path:$model_name"
			} else {
			    set model_unique_name "$model_path:$model_name"
			}
		    } else {
			# for non-hbase storage module, i.e. EDM, use only the name as unique_model_name
			set model_unique_name "$model_name"
		    }
		}
	    }
	}
	# at the moment no GetThumbnailOfModel query
	# Debug "VQuery GetThumbnailOfModel( $model_unique_name)"
	# after 500 [ list VELaSSCo::DoUpdateThumbImage $wThumb [_ "Showing thumbnail for model '%s'" $model_unique_name]]
	set ::VELaSSCo(SelectedModel) $model_unique_name
    }

    proc FillModelTable { tbl} {
	variable _sessionID
	variable _cache

	$tbl delete 0 end

	set group $::VELaSSCo(group_qualifier)
	set name $::VELaSSCo(name_pattern)
	set ret [ gid_velassco GetListOfModels $_sessionID $group $name]
	# Debug "gid_velassco GetListOfModels $_sessionID $group $name = \n$ret"

	# filling list
	set lst_names {}
	set lst_fp {}
	set lst_model_ids {}
	set lst_locations {}
	
	# we skip the first line: "NumberOfModels: N"
	set lst_models [ lrange [ split $ret "\n"] 1 end]
	foreach entry $lst_models {
	    set keyword [ lindex $entry 0]
	    set value [ lrange $entry 1 end]
	    switch $keyword {
		"Name:" {
		    lappend lst_names $value
		}
		"FullPath:" {
		    lappend lst_fp $value
		}
		"ModelID:" {
		    lappend lst_model_ids $value
		}
		"Location:" {
		    lappend lst_locations $value
		}
	    }
	}
	set _cache(Models,Names) {}
	foreach name $lst_names fp $lst_fp lc $lst_locations modelID $lst_model_ids {
	    $tbl insert end [ list $name $fp $lc $modelID]
	    # after 100
	    # set full_name ${name}::${fp}
	    # there will be repeated names ...
	    set full_name "${name} ( ${lc})"
	    lappend _cache(Models,Names) $full_name
	    set _cache(Models,Details,Location,$full_name) $lc
	    set _cache(Models,Details,ModelID,$full_name) $modelID
	    set _cache(Models,Details,FullPath,$full_name) $fp
	    set _cache(Models,Details,Name,$full_name) $name
	}
    }

    proc DoDiscrete2Continuum { w} {
	variable _sessionID
	variable _modelID
	variable _cache
	variable _d2c

	if { ![ CheckLoginAndModelID]} { return}
	WaitState

	# gid_velassco GetDiscrete2ContinuumOfAModel sessionID modelID meshName analysisName 
	#     StepSelectionMode={All,Single or Interval} StepValuesList 
	#     CoarseGrainingMethod={Heaviside or Gaussian} CoarseGrainingWidth CoarseGrainingCutOff 
	#     doProcessContacts 
	#     doTemporalAveraging TemporalAvegaringMode={All or TemporalWindow} DeltaTforTemporalWindow

	set lst_steps {}
	switch [ string tolower $_d2c(StepOptions,SelectionMode)] {
	    "all" {
		set lst_steps $_cache(ListOfSteps)
	    }
	    "single" {
		set lst_steps [ list $_d2c(StepOptions,Step1)]
	    }
	    "interval" {
		set lst_steps [ list $_d2c(StepOptions,Step1) $_d2c(StepOptions,Step2)]
	    }
	}

	set cmd [ list gid_velassco GetDiscrete2ContinuumOfAModel $_sessionID $_modelID $_d2c(StaticMeshHexID) $_d2c(AnalysisName) \
		      $_d2c(StepOptions,SelectionMode) $lst_steps \
		      $_d2c(GrainingOptions,Method) $_d2c(GrainingOptions,Width) $_d2c(GrainingOptions,CutOffFactor) \
		      $_d2c(ProcessContacts,Enabled) \
		      $_d2c(TemporalAverage,Enabled) $_d2c(TemporalAverage,Mode) $_d2c(TemporalAverage,DeltaT)]
	
	Verbose 1 "calling $cmd"
	W "calling $cmd"
	set d2c_result ""
	# set err [ catch {
	#     # set d2c_result [ {*}$cmd]
	# } err_txt]
	# if { $err} {
	#     W "    VQuery error: $err_txt"
	# }
	# set modelBBox [ list 0.0 0.0 0.0 1.0 1.0 1.0]
	Verbose 1 "    result d2c = $d2c_result"
	W "    result d2c = $d2c_result"
	
	EndWaitState
    }

    proc ChangeEnableStatus { var_name lst_w} {
	upvar #0 $var_name value
	set state disabled
	if { $value} {
	    set state normal
	}
	foreach w $lst_w {
	    set err [ catch {
		$w configure -state $state
	    } err_txt]
	    if { $err} {
		W "Couldn't ChangeEnableStatus of $w = [ winfo class $w] to '$state'"
		W "         $err_txt"
	    }
	}
    }

    proc ChangeStepOptions { var_name l_step1 e_step1 l_step2 e_step2} {
	upvar #0 $var_name value
	switch $value {
	    "ALL" {
		$l_step1 configure -state disabled
		$e_step1 configure -state disabled
		$l_step2 configure -state disabled
		$e_step2 configure -state disabled
	    }
	    "SINGLE" {
		$l_step1 configure -state normal
		$e_step1 configure -state normal
		$l_step2 configure -state disabled
		$e_step2 configure -state disabled
		$l_step1 configure -text [_ "step value:"]
		$l_step2 configure -text ""
	    }
	    "INTERVAL" {
		$l_step1 configure -state normal
		$e_step1 configure -state normal
		$l_step2 configure -state normal
		$e_step2 configure -state normal
		$l_step1 configure -text [_ "starting step:"]
		$l_step2 configure -text [_ "ending step:"]
	    }
	    "SET" {
	    }
	}
    }

    proc D2C_window { { w .gid.wVelasscoDiscrete2Continuum}} {
	variable _selected
	variable _cache
	variable _d2c
	if { ![ CheckLoginAndModelID]} { return}
	# check if we have a DEM model
	if { ![ info exists _selected(analysis)] || ![ info exists _cache(InGiD,DynamicMesh)]} {
	    WarnWin [_ "Please, get the model information first."]
	    return
	}
	if { ( $_selected(analysis) != "DEM") || !$_cache(InGiD,DynamicMesh)} {
	    WarnWin [_ "Please, load a DEM model."]
	    return
	}

	# create window
	if { [ winfo exists $w]} {
	    destroy $w
	}
	toplevel $w
	wm title $w [_ "VELaSSCo Discrete 2 Continuum ..."]
	CreateDecorationWindow $w  $w.top $w.bot\
	    [ list [_ "Do D2C"] [_ "Cancel"]] \
	    [ list [ list VELaSSCo::DoDiscrete2Continuum $w] [ list destroy $w]] \
	    [ list <Return> <Escape>]
	ttk::frame $w.top.header
	ttk::label $w.top.header.l1 -image [ VELaSSCo::GetImage VelasscoBig67x48.png]
	# ttk::label $w.top.header.l2 -text [_ "VELaSSCo Discrete 2 Continuum transformation"]
	label $w.top.header.l2 -text [_ "VELaSSCo \nDiscrete 2 Continuum transformation"]
	grid $w.top.header.l1 $w.top.header.l2 -padx 4 -pady 4 -sticky ew

	# ttk::labelframe $w.top.d2c_opt -text [_ "Options"]
	ttk::frame $w.top.d2c_opt
	# Static mesh selection
	# ttk::label $w.top.d2c_opt.l_static -text [_ "Static mesh ID (Hexadecimal string):"]
	ttk::label $w.top.d2c_opt.l_static -text [_ "Model name with static mesh:"]
	# ttk::entry $w.top.d2c_opt.e_static -width 32 -textvariable ::VELaSSCo::_d2c(StaticMeshHexID)
	set lst_labels $_cache(Models,Names)
	set lst_values {}
	foreach full_name $_cache(Models,Names) {
	    lappend lst_values $_cache(Models,Details,ModelID,$full_name)
	}
	TTKMenubutton $w.top.d2c_opt.e_static \
	    -labels $lst_labels \
            -values $lst_values \
            -textvariable ::VELaSSCo::_d2c(StaticMeshHexID)
	# -modifycmd
	# ::VELaSSCo::_d2c(StaticMeshHexID) is used to store the selected ModelID with the static mesh
	# lst_values have the lsit of modelID's
	
	# Analysis selection
	ttk::label $w.top.d2c_opt.l_analysis -text [_ "Output D2C analysis name:"]
	# TTKMenubutton $w.top.d2c_opt.mb_analysis -labels $_cache(ListOfAnalyses) \
        #     -values $_cache(ListOfAnalyses) \
        #     -textvariable ::VELaSSCo::_d2c(AnalysisName)
	ttk::entry $w.top.d2c_opt.mb_analysis -width 16 -textvariable ::VELaSSCo::_d2c(AnalysisName)
	
	# Step options
	ttk::labelframe $w.top.d2c_opt.step_opt -text [_ "Step selection "]
	ttk::label $w.top.d2c_opt.step_opt.l_mode -text [_ "Selection mode:"]
	set sel_mode_labels [ list [_ All] [_ Single] [_ Interval]]
	set sel_mode_values [ list ALL SINGLE INTERVAL]
	TTKMenubutton $w.top.d2c_opt.step_opt.mb_mode \
	    -labels $sel_mode_labels \
            -values $sel_mode_values \
            -textvariable ::VELaSSCo::_d2c(StepOptions,SelectionMode) \
	    -modifycmd [ list ::VELaSSCo::ChangeStepOptions ::VELaSSCo::_d2c(StepOptions,SelectionMode) \
			     $w.top.d2c_opt.step_opt.l_step1 \
			     $w.top.d2c_opt.step_opt.e_step1 \
			     $w.top.d2c_opt.step_opt.l_step2 \
			     $w.top.d2c_opt.step_opt.e_step2 ]
	#-modifycmd [ list ChangeEnableStatus ::VELaSSCo::_d2c(StepOptions,SelectionMode) [ list ]]
	# at the moment: 2 step values
	# if sel_mode == Single --> enable step1, text = step value
	# if sel_mode == Interval -->  enable step1 & step2, text = starting step & ending step
	ttk::label $w.top.d2c_opt.step_opt.l_step1 -text [_ "starting step:"]
	# eventually to be changed to ComboBox
	ttk::entry $w.top.d2c_opt.step_opt.e_step1 -width 8 -textvariable ::VELaSSCo::_d2c(StepOptions,Step1)
	ttk::label $w.top.d2c_opt.step_opt.l_step2 -text [_ "ending step:"]
	# eventually to be changed to ComboBox
	ttk::entry $w.top.d2c_opt.step_opt.e_step2 -width 8 -textvariable ::VELaSSCo::_d2c(StepOptions,Step2)

	grid $w.top.d2c_opt.step_opt.l_mode $w.top.d2c_opt.step_opt.mb_mode
	grid configure $w.top.d2c_opt.step_opt.l_mode  -sticky e
	grid configure  $w.top.d2c_opt.step_opt.mb_mode -sticky w
	grid $w.top.d2c_opt.step_opt.l_step1 $w.top.d2c_opt.step_opt.e_step1 -sticky e -padx 2
	grid configure  $w.top.d2c_opt.step_opt.e_step1 -sticky ew
	grid $w.top.d2c_opt.step_opt.l_step2 $w.top.d2c_opt.step_opt.e_step2 -sticky e -padx 2
	grid configure  $w.top.d2c_opt.step_opt.e_step2 -sticky ew
	
	# Coarse graining...
	ttk::labelframe $w.top.d2c_opt.graining_opt -text [_ "Coarse graining "]
	#eventually change to radiobuttons ...
	ttk::label $w.top.d2c_opt.graining_opt.l_method -text [_ "Method:"]
	set sel_method_labels [ list [_ Heaviside] [_ Gaussian]]
	set sel_method_values [ list Heaviside Gaussian]

	proc ::VELaSSCo::l_UpdateD2CMethodImage { w} {
	    $w configure -image [ VELaSSCo::GetImage d2c_${::VELaSSCo::_d2c(GrainingOptions,Method)}.png]
	}
	TTKMenubutton $w.top.d2c_opt.graining_opt.mb_method \
	    -labels $sel_method_labels \
            -values $sel_method_values \
            -textvariable ::VELaSSCo::_d2c(GrainingOptions,Method) \
 	    -modifycmd "::VELaSSCo::l_UpdateD2CMethodImage $w.top.d2c_opt.graining_opt.img_method"
	ttk::label $w.top.d2c_opt.graining_opt.img_method

	ttk::label $w.top.d2c_opt.graining_opt.l_width -text [_ "Width"]
	# eventually to be changed to ComboBox
	ttk::entry $w.top.d2c_opt.graining_opt.e_width -width 8 -textvariable ::VELaSSCo::_d2c(GrainingOptions,Width)
	ttk::label $w.top.d2c_opt.graining_opt.l_cutoff -text [_ "Cut-off factor"]
	# eventually to be changed to ComboBox
	ttk::entry $w.top.d2c_opt.graining_opt.e_cutoff -width 8 -textvariable ::VELaSSCo::_d2c(GrainingOptions,CutOffFactor)

	grid $w.top.d2c_opt.graining_opt.l_method $w.top.d2c_opt.graining_opt.mb_method
	grid configure $w.top.d2c_opt.graining_opt.l_method  -sticky e
	grid configure  $w.top.d2c_opt.graining_opt.mb_method -sticky w
	grid $w.top.d2c_opt.graining_opt.img_method -columnspan 2 -sticky ew
	grid $w.top.d2c_opt.graining_opt.l_width $w.top.d2c_opt.graining_opt.e_width -sticky e -padx 2
	grid configure  $w.top.d2c_opt.graining_opt.e_width -sticky w
	grid $w.top.d2c_opt.graining_opt.l_cutoff $w.top.d2c_opt.graining_opt.e_cutoff -sticky e -padx 2
	grid configure  $w.top.d2c_opt.graining_opt.e_cutoff -sticky w

	# Process contacts
	ttk::checkbutton $w.top.d2c_opt.cb_contacts -text [_ "Process contacts"] \
	    -variable ::VELaSSCo::_d2c(ProcessContacts,Enabled)

	# Temporal Averaging
	ttk::labelframe $w.top.d2c_opt.temp_avg_opt -text [_ "Temporal averaging "]
	ttk::checkbutton $w.top.d2c_opt.temp_avg_opt.cb_do -text [_ "Do Temporal averaging"] \
	    -variable ::VELaSSCo::_d2c(TemporalAverage,Enabled) \
	    -command [ list ::VELaSSCo::ChangeEnableStatus ::VELaSSCo::_d2c(TemporalAverage,Enabled) \
			   [ list  $w.top.d2c_opt.temp_avg_opt.cb_opt1  $w.top.d2c_opt.temp_avg_opt.cb_opt2 \
				 $w.top.d2c_opt.temp_avg_opt.l_deltaT $w.top.d2c_opt.temp_avg_opt.e_deltaT]]
	ttk::radiobutton $w.top.d2c_opt.temp_avg_opt.cb_opt1 \
	    -text [_ All] -value ALL -variable ::VELaSSCo::_d2c(TemporalAverage,Mode)
	ttk::radiobutton $w.top.d2c_opt.temp_avg_opt.cb_opt2 \
	    -text [_ "Temporal window"] -value TemporalWindow -variable ::VELaSSCo::_d2c(TemporalAverage,Mode)
	ttk::label $w.top.d2c_opt.temp_avg_opt.l_deltaT -text [_ "DeltaT"]
	ttk::entry $w.top.d2c_opt.temp_avg_opt.e_deltaT -width 8 -textvariable ::VELaSSCo::_d2c(TemporalAverage,DeltaT)
	
	grid $w.top.d2c_opt.temp_avg_opt.cb_do -sticky w -columnspan 2 -padx 4
	grid $w.top.d2c_opt.temp_avg_opt.cb_opt1 -sticky w -columnspan 2 -padx 16
	grid $w.top.d2c_opt.temp_avg_opt.cb_opt2 -sticky w -columnspan 2 -padx 16
	grid $w.top.d2c_opt.temp_avg_opt.l_deltaT $w.top.d2c_opt.temp_avg_opt.e_deltaT -sticky e -padx 16
	grid configure  $w.top.d2c_opt.temp_avg_opt.e_deltaT -sticky w -padx 0

	# pack frames:
	grid $w.top.d2c_opt.l_analysis $w.top.d2c_opt.mb_analysis  -padx 4 -pady 4 -sticky e
	# grid $w.top.d2c_opt.l_static $w.top.d2c_opt.e_static -padx 4 -pady 4 -sticky e
	# grid configure  $w.top.d2c_opt.e_static -sticky w
	grid $w.top.d2c_opt.l_static -sticky w -columnspan 2 -padx 4 -pady 4 
	grid $w.top.d2c_opt.e_static -sticky e -columnspan 2 -padx 4 -pady 4
	grid $w.top.d2c_opt.step_opt -sticky news -columnspan 2 -padx 4 -pady 4
	grid $w.top.d2c_opt.graining_opt -sticky news -columnspan 2  -padx 4 -pady 4 -sticky news
	grid $w.top.d2c_opt.cb_contacts -sticky ew -columnspan 2  -padx 4 -pady 4 -sticky news
	grid $w.top.d2c_opt.temp_avg_opt -sticky news -columnspan 2  -padx 4 -pady 4 -sticky news

	grid configure $w.top.d2c_opt.mb_analysis  -sticky w

	grid $w.top.header -sticky ew
	grid $w.top.d2c_opt -sticky ewns
	grid columnconfigure $w.top 0 -weight 1
	grid rowconfigure $w.top 1 -weight 1


	# Default options:
	if { ![ info exists _d2c(AnalysisName)]} {
	    # set _d2c(StaticMeshHexID) "708f23d401f17c34ac3df5bff499032b"
	    # set _d2c(OutputModelName) "D2C_Mesh"
	    set _d2c(StaticMeshHexID) "D2C_Mesh"
	    set _d2c(AnalysisName) $_selected(analysis) 
	    set _d2c(StepOptions,SelectionMode) INTERVAL
	    set _d2c(StepOptions,Step1) [ lindex $_cache(ListOfSteps) 0]
	    set _d2c(StepOptions,Step2) [ lindex $_cache(ListOfSteps) end]
	    set _d2c(GrainingOptions,Method) Gaussian
	    $w.top.d2c_opt.graining_opt.img_method configure \
		-image [ VELaSSCo::GetImage d2c_${::VELaSSCo::_d2c(GrainingOptions,Method)}.png]
	    set _d2c(GrainingOptions,Width) 0.0024
	    set _d2c(GrainingOptions,CutOffFactor) 3
	    set _d2c(ProcessContacts,Enabled) 1
	    set _d2c(TemporalAverage,Enabled) 1
	    set _d2c(TemporalAverage,Mode) ALL
	    set _d2c(TemporalAverage,DeltaT) 30000
	}
    }

    proc SelectModel { { w .gid.wVelasscoSelectModel}} {
	if { ![ CheckLogin]} { return}
	if { [ winfo exists $w]} {
	    destroy $w
	}
	toplevel $w
	wm title $w [_ "VELaSSCo model selection..."]
	# CreateApplyCloseDecorationWindow $w $w.top $w.bot \
	    # 	{WarnWin Accept} \
	    # 	"destroy $w"
	CreateDecorationWindow $w  $w.top $w.bot\
	    [ list [_ "Open"] [_ "Cancel"]] \
	    [ list [ list VELaSSCo::DoSelectModel $w] [ list destroy $w]] \
	    [ list <Return> <Escape>]
	ttk::frame $w.top.header
	ttk::label $w.top.header.l1 -image [ VELaSSCo::GetImage VelasscoBig67x48.png]
	ttk::label $w.top.header.l2 -text [_ "VELaSSCo model selection"]
	grid $w.top.header.l1 $w.top.header.l2 -padx 4 -pady 4 -sticky ew

	package require tablelist_tile
	ttk::frame $w.top.selection

	ttk::frame $w.top.selection.filter
	ttk::label $w.top.selection.filter.lgroup -text [_ "Group qualifier:"]
	ttk::entry $w.top.selection.filter.egroup -width 20 -textvariable ::VELaSSCo(group_qualifier)
	ttk::label $w.top.selection.filter.lpattern -text [_ "Name pattern:"]
	ttk::entry $w.top.selection.filter.epattern -width 20 -textvariable ::VELaSSCo(name_pattern)
	ttk::button $w.top.selection.filter.brefresh -text [_ "Refresh table"]
	grid $w.top.selection.filter.lgroup $w.top.selection.filter.egroup \
	     $w.top.selection.filter.lpattern $w.top.selection.filter.epattern \
	    $w.top.selection.filter.brefresh \
	    -padx 4 -pady 4 -sticky e
	grid configure  $w.top.selection.filter.egroup -sticky w
	grid configure  $w.top.selection.filter.epattern -sticky w
	grid configure  $w.top.selection.filter.brefresh -sticky w

	set tbl_sw [ ScrolledWindow $w.top.selection.sw -auto both -relief solid -borderwidth 0]
	set tbl $w.top.selection.sw.tbl
	set strech_opt all
	# strech Full path and location
	set strech_opt [ list 1 2]
	tablelist::tablelist $tbl \
	    -stretch $strech_opt -background white \
	    -columns [ list 20 [_ "Name"] 30 [_ "Full path"] 15 [_ "Location"] 20 [_ "Model ID"]]
	    # -columns [ list 30 [_ "Name"] 0 [_ "Full path"] 20 [_ "Validation status"] 20 [_ "Access"]]

	$w.top.selection.sw setwidget $tbl

	ttk::label $w.top.selection.thumb -text [_ "Thumbnail image of selected model"]
	bind $tbl <<TablelistSelect>> [ list VELaSSCo::GetAndUpdateThumbnail %x %y %W $tbl $w.top.selection.thumb]

	bind [$tbl bodytag] <Double-Button-1> [ list VELaSSCo::DoSelectModel $w]

	grid $w.top.selection.filter -sticky ew
	# grid $w.top.selection.tbl -padx 4 -pady 4 -sticky news
	grid $w.top.selection.sw -padx 4 -pady 4 -sticky news
	grid $w.top.selection.thumb  -padx 4 -pady 4 -sticky ew

	grid columnconfigure $w.top.selection 0 -weight 1
	grid rowconfigure $w.top.selection 1 -weight 1

	grid $w.top.header -sticky ew
	grid $w.top.selection -sticky ewns
	grid columnconfigure $w.top 0 -weight 1
	grid rowconfigure $w.top 1 -weight 1
	wm geometry $w 990x325

	if { ![ info exists ::VELaSSCo(group_qualifier)]} {
	    set ::VELaSSCo(group_qualifier) "*"
	    set ::VELaSSCo(name_pattern) "*"
	}

	$w.top.selection.filter.brefresh configure -command [ list ::VELaSSCo::FillModelTable $tbl] 
	FillModelTable $tbl
    }

    proc AlreadyInGiD { selected_analysis selected_step res_name} {
	variable _cache
	set in_gid 0
	if { !$_cache(InGiD,DynamicMesh)  && \
		 [ info exists _cache(InGiD,Result,$selected_analysis,$selected_step,$res_name)] && \
		 $_cache(InGiD,Result,$selected_analysis,$selected_step,$res_name)} {
	    set in_gid 1
	}
	return $in_gid
    }

    proc ExpandResultsInCurrentStep { selected_analysis selected_step res_name lst_ids } {
	variable _cache
	variable _dummy_result_value
	if { ![ info exist _cache($selected_analysis,$selected_step,$res_name,Header)]} {
	    # may be is a result in gauss points which are not handled at the moment
	    return
	}
	set res_header $_cache($selected_analysis,$selected_step,$res_name,Header)
	set res_comp $_cache($selected_analysis,$selected_step,$res_name,ComponentNames)
	set res_type [ string tolower [ lindex $res_header 4]]
	set location [ string tolower [ lindex $res_header 5]]
	set num_comp $_cache($selected_analysis,$selected_step,$res_name,NumberOfComponents)
	
	# define dummy result values for current loaded mesh
	set res_type2 [ string tolower $_cache($selected_analysis,$selected_step,$res_name,Type)]
	if { $res_type != $res_type2} {
	    ErrorWin "$res_type != $res_type2"
	    return
	}

	if { $location != "onnodes"} {
	    return
	}
	# id_node result_value(s)
	set lst_gid_res {}
	foreach v_id $lst_ids {
	    set dummy_result [ list $v_id $_dummy_result_value]
	    switch $res_type {
		scalar { set dummy_result [ list $v_id $_dummy_result_value]}
		vector { set dummy_result [ list $v_id \
						$_dummy_result_value $_dummy_result_value $_dummy_result_value]}
		matrix { set dummy_result [ list $v_id \
						$_dummy_result_value $_dummy_result_value $_dummy_result_value \
						$_dummy_result_value $_dummy_result_value $_dummy_result_value]}
	    }
	    lappend lst_gid_res $dummy_result
	}
	
	# set lst_gid_res [ list $lst_ids $lst_values]
	if { [ llength $res_comp] == 0} {
	    GiD_Result create $res_header {*}$lst_gid_res
	} else {
	    set components [ list ComponentNames {*}$res_comp]
	    GiD_Result create $res_header $components {*}$lst_gid_res
	}
    }
    
    proc CreateResultInGiD { selected_analysis selected_step res_name lst_ids lst_values} {
	variable _cache
	# res_header = [ list Result $res_name $selected_analysis $selected_step $res_type $location]
	set res_header $_cache($selected_analysis,$selected_step,$res_name,Header)
	set res_comp $_cache($selected_analysis,$selected_step,$res_name,ComponentNames)
	set res_type [ lindex $res_header 4]
	set num_comp $_cache($selected_analysis,$selected_step,$res_name,NumberOfComponents)
	
	set lst_gid_res {}
	for { set idx_id 0} { $idx_id < [ llength $lst_ids]} { incr idx_id} {
	    set node_id [ lindex $lst_ids $idx_id]
	    set value [ lindex $lst_values $idx_id]
	    lappend lst_gid_res [ list $node_id {*}$value]
	}
	# set lst_gid_res [ list $lst_ids $lst_values]
	if { [ llength $res_comp] == 0} {
	    GiD_Result create $res_header {*}$lst_gid_res
	} else {
	    set components [ list ComponentNames {*}$res_comp]
	    GiD_Result create $res_header $components {*}$lst_gid_res
	}
	if { [ llength $lst_gid_res] > 0} {
	    set _cache(InGiD,Result,$selected_analysis,$selected_step,$res_name) 1
	}
    } 

    proc CreateResultTree { selected_analysis lst_steps lst_results} {
	variable _selected
	variable _cache
	variable _dummy_mesh_name
	# assuming the results are the same for all time-steps:
	set lst_results [ split $lst_results \n]
	set num_res 0
	set read_num_res 0
	set _cache(ListOfResults,$selected_analysis) {}
	foreach pair $lst_results {
	    set key [ lindex $pair 0]
	    if { $key == ""} {
		continue
	    }
	    switch $key {
		"NumberOfResults:" {
		    set read_num_res [ lindex $pair 1]
		}
		"Name:" {
		    # new result
		    set resInfo($num_res,$key) [ lrange $pair 1 end]
		    lappend _cache(ListOfResults,$selected_analysis) [ lrange $pair 1 end]
		}
		"NumberOfComponents:" {
		    set resInfo($num_res,$key) [ lrange $pair 1 end]
		}
		"ComponentNames:" {
		    set resInfo($num_res,$key) [ lrange $pair 1 end]
		}
		"ResultNumber:" {
		    # last result info field
		    set resInfo($num_res,$key) [ lindex $pair 1]
		    incr num_res
		}
		default {
		    set resInfo($num_res,$key) [ lrange $pair 1 end]
		}
	    }
	}
	
	# lets create dummy results in order to have the result menu's...
	# GiD_Result needs a mesh first ...
	# GiD_MeshPost create "dummyMesh_*back*" Point 1 {1} {0.0 0.0 0.0} {1} {1}
	# GiD_MeshPost create "dummyMesh_*back*" Sphere 1 {1} {0.0 0.0 0.0} {1} {1} {0.0}

	set num_steps [ llength $lst_steps]
	set total_results_to_create [ expr $num_steps * $num_res]
	set use_file_res 0
	set file_res_data {}
	set use_file_msh 0
	set file_msh_data {}
	# inside GiD there is a quadratic loop, so to create lots of results
	# we'll create a temporary file and read it in GiD
	if { ( $total_results_to_create > 10000) && ( $num_steps > 1000)} {
	    # set use_file_res 1
	    # lappend file_res_data "GiD Post Results File 1.1"
	    # lappend file_res_data "# encoding utf-8"
	    # lappend file_res_data {}

	    # trick for the use case:
	    # select the first 50, the last 50 and the 50 just after step = 2939000 (included)
	    set original_lst_steps $lst_steps
	    # find step  = 2939000
	    set idx_use_case_step [ lsearch $original_lst_steps 2939000.0]
	    if {  $idx_use_case_step < 0} {
		# i.e. not found
		set lst_steps [ lrange $original_lst_steps 0 49]
		lappend lst_steps {*}[ lrange $original_lst_steps end-49 end]
	    } else {
		# first 50
		set lst_steps [ lrange $original_lst_steps 0 49]
		set end_idx_first_50 49
		set ini_idx_last_50 [ expr $num_steps - 50]
		if { $idx_use_case_step < $end_idx_first_50} {
		    # use case step is inside, add some more until 50 after use-case step
		    set end_idx_first_50 [ expr $idx_use_case_step + 49]
		    lappend lst_steps {*}[ lrange $original_lst_steps 50 $end_idx_first_50]
		} else {
		    # add 50 from use-case step
		    set end_idx_first_50 [ expr $idx_use_case_step + 49]
		    lappend lst_steps {*}[ lrange $original_lst_steps $idx_use_case_step $end_idx_first_50]
		}
		if { $end_idx_first_50 > $ini_idx_last_50} {
		    # end of 50 after current use-case is inside last 50, add rest
		    lappend lst_steps {*}[lrange  $original_lst_steps $end_idx_first_50 end]
		} else {
		    lappend lst_steps {*}[lrange  $original_lst_steps $ini_idx_last_50 end]
		}
	    }
	    # W "Creating results for these time-steps:"
	    # W "    $lst_steps"
	}

	if { [ llength [ array names _cache InGiD,Mesh,*]] == 0} { # first time ever...
	    if { !$use_file_res} {
		GiD_MeshPost create $_dummy_mesh_name Sphere 1 -zero_based_array {0} {0.0 0.0 0.0} {0} {0} {0.0}
	    } else {
		lappend file_msh_data "# encoding utf-8"
		lappend file_msh_data {}
		lappend file_msh_data "MESH \"$_dummy_mesh_name\" dimension 3 ElemType Sphere Nnode 1"
		lappend file_msh_data "# Color 0.600000 0.600000 0.600000"
		lappend file_msh_data "Coordinates"
		lappend file_msh_data "1 0.0 0.0 0.0"
		lappend file_msh_data "End coordinates"
		lappend file_msh_data "Elements"
		lappend file_msh_data "1 1 0.0"
		lappend file_msh_data "End elements"
		lappend file_msh_data {}
	    }
	    set use_file_msh $use_file_res
	    set _cache(InGiD,Mesh,$_dummy_mesh_name) $_dummy_mesh_name
	    set _cache(VerticesID) {1}
	    set _cache(InGiD,MeshElemType,$_dummy_mesh_name) Sphere
	}

	# problems with:
	# GiD_MeshPost create dummy_mesh Sphere 1 -zero_based_array {0} {0.0 0.0 0.0} {0} {0} {1.0}
	# GiD_Result create {Result {Element's centre} geometry 1.0 vector OnGaussPoints GP_SPHERE_1} {ComponentNames {X-Element's centre} {Y-Element's centre} {Z-Element's centre} {|Element's centre|}} {1 0 0 0}
	foreach selected_step $lst_steps {
	    for { set ires 0} { $ires < $read_num_res} { incr ires} {
		set res_name $resInfo($ires,Name:)
		# analysis already selected
		# step already selected, eventually loop over all steps to create the results
		# some times type begins with uppercase, sometimes not!
		set res_type [ string tolower $resInfo($ires,ResultType:)]
		# id_node result_value(s)
		set dummy_result {1 0}
		switch $res_type {
		    scalar { set dummy_result {1 0}}
		    vector { set dummy_result {1 0 0 0}}
		    matrix { set dummy_result {1 0 0 0 0 0 0}}
		}
		set location $resInfo($ires,Location:)
		set res_header [ list Result $res_name $selected_analysis $selected_step $res_type $location]
		incr ::TeoricalResultsCreated 
		if { [ string equal -nocase $location "OnGaussPoints"]} {
		    set gp_name $resInfo($ires,GaussPointName:)
		    set lc_gp_name [ string tolower $resInfo($ires,GaussPointName:)]
		    # if it begins with GP_ and ends with _1, like GP_TETRA.._1 GP_SPHERE_1
		    # change it to GP_ELELEMENT_1
		    if { [ regexp {^gp_.*_1$} $lc_gp_name] == 1} {
			set gp_name GP_SPHERE_1
		    }
		    lappend res_header $gp_name
		    # do not create gauss point results
		    if { ![ info exist ::VELaSSCo::_warning_gp_once] || $::VELaSSCo::_warning_gp_once} {
			set ::VELaSSCo::_warning_gp_once 0
			W "Warning: At the moment Gauss Points are not handled"
		    }
		    continue
		}
		if { !$use_file_res} {
		    if { [ llength $resInfo($ires,ComponentNames:)] == 0} {
			set err [ catch {
			    GiD_Result create $res_header $dummy_result
			} err_txt ]
			if { $err} {
			    W "Error creatring GiD_Result create {$res_header} {$dummy_result}"
			    W "        Error = $err_txt"
			}
		    } else {
			set components [ list ComponentNames {*}$resInfo($ires,ComponentNames:)]
			set err [ catch {
			    GiD_Result create $res_header $components $dummy_result
			} err_txt ]
			if { $err} {
			    W "Error creating GiD_Result create {$res_header} {$components} {$dummy_result}"
			    W "        with component names = $resInfo($ires,ComponentNames:)"
			    W "        Error = $err_txt"
			}
		    }
		} else {
		    lappend file_res_data $res_header
		    if { [ llength $resInfo($ires,ComponentNames:)] != 0} {
			set ln "ComponentNames \"[ lindex $resInfo($ires,ComponentNames:) 0]\""
			for { set i 1} { $i < [ llength $resInfo($ires,ComponentNames:)]} { incr i} {
			    append ln ", \"[ lindex $resInfo($ires,ComponentNames:) $i]\""
			}
			lappend file_res_data $ln
		    }
		    lappend file_res_data "Values"
		    lappend file_res_data $dummy_result
		    lappend file_res_data "End values"
		    lappend file_res_data {}
		    
		}
		incr ::NumResultsCreated
		if { ![ expr $::NumResultsCreated % 1000]} {
		    W "   ... $::NumResultsCreated result's entries created"
		}

		# W "created  $res_header"
		# GiD uses 'underscore' to encode spaces, so eventually we've to cache that too
		set _cache($selected_analysis,$selected_step,$res_name,Header) $res_header
		set _cache($selected_analysis,$selected_step,$res_name,Type) $res_type
		set _cache($selected_analysis,$selected_step,$res_name,ComponentNames) $resInfo($ires,ComponentNames:)
		set _cache($selected_analysis,$selected_step,$res_name,NumberOfComponents) $resInfo($ires,NumberOfComponents:)
		regsub -all {_} $selected_analysis { } gid_an
		set _cache($gid_an,$selected_step,$res_name,Header) $res_header
		set _cache($gid_an,$selected_step,$res_name,Type) $res_type
		set _cache($gid_an,$selected_step,$res_name,ComponentNames) $resInfo($ires,ComponentNames:)
		set _cache($gid_an,$selected_step,$res_name,NumberOfComponents) $resInfo($ires,NumberOfComponents:)
		set _cache($gid_an,OriginalName) $selected_analysis
		set _cache($selected_analysis,GiDName) $gid_an
		regsub -all {_} $res_name { } gid_res
		set _cache($gid_res,OriginalName) $res_name
	    }
	}

	# if there are too many results write them in a file and import them in gid:
	if { $use_file_res} {
	    # mergo to existing info
	    set cmd Add
	    set tmp_filename_msh [ file join [ GiD_Info project TmpDirectory] "velassco_tmp_mesh.post.msh"]
	    file delete -force $tmp_filename_msh
	    if { $use_file_msh} {
		# new model
		set cmd Read
		set fo [ open $tmp_filename_msh w]
		puts $fo [ join $file_msh_data \n]
		close $fo
		W "velassco temporal mesh file created: $tmp_filename_msh"
	    }
	    set tmp_filename_res [ file join [ GiD_Info project TmpDirectory] "velassco_tmp_mesh.post.res"]
	    file delete -force $tmp_filename_res
	    set fo [ open $tmp_filename_res w]
	    puts $fo [ join $file_res_data \n]
	    close $fo
	    W "velassco temporal result file created: $tmp_filename_res"

	    GiD_Process Mescape Files $cmd $tmp_filename_res
	    # file delete -force $tmp_filename_msh
	    # file delete -force $tmp_filename_res
	}
    }

    proc GetModelInfo { show_info} {
	variable _sessionID
	variable _modelID
	variable _cache
	variable _selected
	if { ![ CheckLoginAndModelID]} { return}
	set doing ""
	WaitState
	set err [ catch {
	    # testing GetListOfAnalyses
	    if { ![ info exists _cache(ListOfAnalyses)] || ![ info exists _selected(analysis)]} {
		set doing "gid_velassco GetListOfAnalyses $_sessionID $_modelID"
		Verbose 1 $doing
		set lst_analyses [ gid_velassco GetListOfAnalyses $_sessionID $_modelID]
		set selected_analysis [ lindex $lst_analyses 0]
		if { [ llength $lst_analyses] > 1} {
		    # if more than one, select the last one
		    set selected_analysis [ lindex $lst_analyses end]
		}
		set _selected(analysis) $selected_analysis
		set _cache(ListOfAnalyses) $lst_analyses
		if { $show_info} {
		    W "gid_velassco GetListOfAnalyses $_sessionID $_modelID = "
		    W $lst_analyses
		    # W "  selected = $selected_analysis"
		    W \n
		}
	    } else {
		set selected_analysis $_selected(analysis)
	    }

	    # testing GetListOfSteps
	    if { ![ info exists _cache(ListOfSteps)] || ![ info exists _selected(step)]} {
		set doing "gid_velassco GetListOfSteps $_sessionID $_modelID $selected_analysis"
		Verbose 1 $doing
		set lst_steps [ gid_velassco GetListOfSteps $_sessionID $_modelID $selected_analysis]
		set selected_step [ lindex $lst_steps 0]
		if { [ llength $lst_steps] > 1} {
		    # if more than one, select the last one
		    set selected_step [ lindex $lst_steps end]
		}
		set _selected(step) $selected_step
		set _cache(ListOfSteps) $lst_steps
		set num_steps [ llength $lst_steps]
		if { $show_info} {
		    W "gid_velassco GetListOfSteps $_sessionID $_modelID $selected_analysis = "
		    W "Number of Steps = $num_steps"
		    if { $num_steps < 20} {
			W $lst_steps
		    } else {
			# show only some of them
			W "[ lrange $lst_steps 0 7] . . . [ lrange $lst_steps [ expr $num_steps - 8] end]"
		    }
		    # W "  selected = $selected_step"
		    W \n
		}
	    } else {
		set selected_step $_selected(step)
		set lst_steps $_cache(ListOfSteps)
	    }

	    # testing GetListOfResults
	    if { ![ info exists _cache(ListOfResults,$selected_analysis)]} {
		set doing "gid_velassco GetListOfResults $_sessionID $_modelID $selected_analysis $selected_step"
		Verbose 1 $doing
		set lst_results [ gid_velassco GetListOfResults $_sessionID $_modelID $selected_analysis $selected_step]
		if { $show_info} {
		    W "gid_velassco GetListOfResults $_sessionID $_modelID $selected_analysis $selected_step = "
		    W $lst_results
		    W \n
		}
		set _cache(ListOfResults,$selected_analysis) $lst_results
	    } else {
		set lst_results $_cache(ListOfResults,$selected_analysis)
	    }

	    # testing GetListOfMeshes
	    if { ![ info exists _cache(GetListOfMeshes,Results)] } {
		set gm_analysis $selected_analysis
		set gm_step $selected_step
		if { "$selected_analysis" != "DEM"} {
		    # get information about static meshes
		    set gm_analysis ""
		    set gm_step 0.0
		}
		set doing "gid_velassco GetListOfMeshes $_sessionID $_modelID $gm_analysis $gm_step"
		Verbose 1 $doing
		set lst_meshes [ gid_velassco GetListOfMeshes $_sessionID $_modelID $gm_analysis $gm_step]
		if { $show_info} {
		    W "gid_velassco GetListOfMeshes $_sessionID $_modelID \"$gm_analysis\" $gm_step = "
		    W $lst_meshes
		    W \n
		}
		
		set _cache(GetListOfMeshes,Query) "GetListOfMeshes $_sessionID $_modelID $gm_analysis $gm_step"
		set _cache(GetListOfMeshes,Results) $lst_meshes
	    }
	} err_txt]

	EndWaitState
	
	if { $err} {
	    W "ERROR doing $doing"
	    W $err_txt
	} else {
	    if { "$selected_analysis" == "DEM"} {
		#dynamic mesh
		set _cache(InGiD,DynamicMesh) 1
	    } else {
		# static mesh
		set _cache(InGiD,DynamicMesh) 0
	    }

	    # delete any previous mesh and results info
	    # GiD_Process Mescape Files New Yes escape
	    # CreateResultTree $selected_analysis $selected_step $lst_results
	    set total_num_res 0
	    set last_analysis ""
	    set last_step 0
	    foreach my_analysis $_cache(ListOfAnalyses) {
		set last_analysis $my_analysis
		set my_lst_steps {}
		set my_step ""
		set my_lst_results {}
		W "getting Steps and Results information for analysis '$my_analysis'..."
		set err [ catch {
			     set my_lst_steps [ gid_velassco GetListOfSteps $_sessionID $_modelID $my_analysis]
			     set my_step [ lindex $my_lst_steps end]
			     set my_lst_results [ gid_velassco GetListOfResults $_sessionID $_modelID $my_analysis $my_step]
			 } err_txt ]
		set last_step $my_step
		if { !$err} {
		    set num_res_step [ llength [ lsearch -all -nocase $my_lst_results "name:"]]
		    set num_res [ expr [ llength $my_lst_steps] * $num_res_step]
		    # W "   CreateResultTree $num_res for $my_analysis: n_steps * n_res = [ llength $my_lst_steps] * [ llength  $my_lst_results]"
set ::NumResultsCreated 0
set ::TeoricalResultsCreated 0
		    CreateResultTree $my_analysis $my_lst_steps $my_lst_results
		    # incr total_num_res $num_res
		    incr total_num_res $::NumResultsCreated
		    W "   $::NumResultsCreated result's entries created"
		} else {
		    W "ERROR getting Steps and Result information for analysis = $my_analysis"
		    W "      lst_steps = $my_lst_steps"
		    W "      lst_results = $my_lst_results"
		}
	    }

	    W "A total of $total_num_res result's entries have created"

	    # select the last one:
	    set gid_an [ toGiDEncoding $last_analysis]
	    GiD_Process Mescape results analysissel $gid_an $last_step
	    set gid_an [ fromGiDEncoding $gid_an]
	    set _selected(analysis) $_cache($gid_an,OriginalName)
	    set _selected(step) $last_step
	    W "   selected analysis = $_selected(analysis)"
	    W "   selected step = $_selected(step)"

	    set num_vertices 1 
	    #[ llength $_cache(VerticesID)]
	    # W "calculated Results created * NumVertices = $total_num_res * $num_vertices = [ expr $total_num_res * $num_vertices]"
	    # W "     real results creted = $::NumResultsCreated"
	    # W "     teor results creted = $::TeoricalResultsCreated"

	    # if { "$selected_analysis" == "DEM"} {
	    # 	#dynamic mesh
	    # 	set _cache(InGiD,DynamicMesh) 1
	    # 	# DEM meshes will be created with the event "LoadSelection" raised by "GiD_Process Mescape results analysissel"
	    # 	GIDcreateDEMMeshes
	    # } else {
	    # 	# static mesh
	    # 	set _cache(InGiD,DynamicMesh) 0
	    # }
	}
    }

    proc DoTunnelLinux {} {

	set velassco_host "$::VELaSSCo(Username)@$::VELaSSCo(Host) port $::VELaSSCo(Port) (tunneling)"
	# set _tunnel_used [ exec ssh -N -L $::VELaSSCo(Tunnel,LocalPort):$::VELaSSCo(Host):$::VELaSSCo(Port) $::VELaSSCo(Tunnel,Username)@$::VELaSSCo(Tunnel,Host) &]
	# needs expect
	set err [ catch {
	    set fo [ file tempfile tmp_file velassco]
	    puts $fo "#!/usr/bin/expect"
	    puts $fo "# Using ssh from expect"
	    puts $fo "log_user 0"
	    puts $fo "spawn ssh -N -L $::VELaSSCo(Tunnel,LocalPort):$::VELaSSCo(Host):$::VELaSSCo(Port) $::VELaSSCo(Tunnel,Username)@$::VELaSSCo(Tunnel,Host)"
	    puts $fo "expect \"sword: \""
	    puts $fo "send \"$::VELaSSCo(Tunnel,Password)\\r\""
	    puts $fo "interact"
	    close $fo
	    file attributes $tmp_file -permissions u+x,go-rwx
	    set _tunnel_used [ exec $tmp_file &]
	    # Debug $_tunnel_used
	    after 1000
	    file delete -force $tmp_file
	} err_txt]
	if { $err} {
	    Print "Could not use expect: '$err_txt', using interactive xterm isntead."
	    set _tunnel_used [ exec xterm -e ssh -N -L $::VELaSSCo(Tunnel,LocalPort):$::VELaSSCo(Host):$::VELaSSCo(Port) $::VELaSSCo(Tunnel,Username)@$::VELaSSCo(Tunnel,Host) &]
	    after 10000
	}
    }

    proc DoTunnelWindows {} {
	variable _dir
	variable _tunnel_used
	
	set velassco_host "$::VELaSSCo(Username)@$::VELaSSCo(Host) port $::VELaSSCo(Port) (tunneling)"
	# set _tunnel_used [ exec ssh -N -L $::VELaSSCo(Tunnel,LocalPort):$::VELaSSCo(Host):$::VELaSSCo(Port) $::VELaSSCo(Tunnel,Username)@$::VELaSSCo(Tunnel,Host) &]
	# needs expect
	set bin_putty [ file join $_dir bin putty.exe]
	set _tunnel_used [ exec "$bin_putty" -N -L $::VELaSSCo(Tunnel,LocalPort):$::VELaSSCo(Host):$::VELaSSCo(Port) $::VELaSSCo(Tunnel,Username)@$::VELaSSCo(Tunnel,Host) &]
	# set putty_cmd "\"$bin_putty\" -N -L $::VELaSSCo(Tunnel,LocalPort):$::VELaSSCo(Host):$::VELaSSCo(Port) $::VELaSSCo(Tunnel,Username)@$::VELaSSCo(Tunnel,Host)"
	# proc ::___dummy__cb {} {
	# }
	# set _tunnel_used [ runExe run $putty_cmd -blocking false -callback ::___dummy__cb -timeout 0]
	after 10000
    }

    proc DoLogin { w} {
	variable _sessionID
	variable _tunnel_used

	if { $_sessionID != ""} {
	    WarnWin [_ "Already logged in!"]
	    return
	}

	if { $::VELaSSCo(Tunnel,enabled)} {

	    if { $::tcl_platform(platform) == "windows"} {
		DoTunnelWindows
	    } else {
		DoTunnelLinux
	    }

	    set gv_id [ gid_velassco UserLogin localhost $::VELaSSCo(Tunnel,LocalPort) $::VELaSSCo(Username) $::VELaSSCo(Password)]

	    # when closing the window do a logout ...
	    bind $w <Destroy> "+::VELaSSCo::DoLogout"

	} else {
	    set _tunnel_used 0
	    set gv_id [ gid_velassco UserLogin $::VELaSSCo(Host) $::VELaSSCo(Port) $::VELaSSCo(Username) $::VELaSSCo(Password)]
	}
	set _sessionID $gv_id

	ShowLoginStatus
    }

    proc DoLogout {} {
	variable _sessionID
	variable _modelID
	variable _tunnel_used
	if { $_sessionID != ""} {
	    if { "$_modelID" != ""} {
		CloseModel
	    }
	    set ret ""
	    catch { 
		set ret [ gid_velassco UserLogout $_sessionID]
	    }
	    if { $_tunnel_used != 0} {
		# kill tunnel
		if { $::tcl_platform(platform) == "windows"} {
		    # runExe stop $_tunnel_used
		    # needed to kill the tunnel
		    package require gid_cross_platform
		    # can not do the package require when needed because sometimes gid is being destroyed
		    gid_cross_platform::end_process $_tunnel_used
		} else {
		    exec kill $_tunnel_used
		}
	    }
	    Verbose 1 "gid_velassco UserLogout $_sessionID = $ret"
	}
	set _sessionID ""
	set _tunnel_used 0
	ShowLoginStatus
    }

    proc DoStatusDB {} {
	variable _sessionID
	if { ![ CheckLogin]} { return}
	InfoWin [_ "gid_velassco GetStatusDB $_sessionID = \n%s" \
		     [ gid_velassco GetStatusDB $_sessionID]]
    }

    proc DoStopVELaSSCo {} {
	variable _sessionID
	if { ![ CheckLogin]} { return}
	InfoWin [_ "gid_velassco StopVELaSSCo $_sessionID = \n%s" \
		     [ gid_velassco StopVELaSSCo $_sessionID]]
	DoLogout
    }

    proc EnableDisableWidgets { lst_widgets var_name} {
	upvar #0 $var_name var_state
	foreach w $lst_widgets {
	    set state disabled
	    if { $var_state} {
		set state normal

		# set new default parameters
		set ::VELaSSCo(Host) pez001
		set ::VELaSSCo(Port) 26267
	    }
	    $w configure -state $state
	}
    }

    proc ShowLoginStatus {} {
	variable _sessionID
	variable _status_wid
	variable _login_icon_wid
	variable _icon_do_connection
	variable _icon_connected
	
	if { ( $_status_wid != "") && [ winfo exists $_status_wid]} {
	    set color green
	    set txt "Logged in"
	    if { $_sessionID == ""} {
		set color red
		set txt "Not logged in"
	    }
	    $_status_wid configure -background $color
	    $_status_wid.txt configure -text $txt -background $color -foreground white
	}

	if { ( $_login_icon_wid != "") && [ winfo exists $_login_icon_wid]} {
	    set img $_icon_connected
	    if { $_sessionID == ""} {
		set img $_icon_do_connection
	    }
	    $_login_icon_wid configure -image [ VELaSSCo::GetImage $img]
	}
    }

    proc CheckLogin {} {
	variable _sessionID

	set ret 1
	if { $_sessionID == ""} {
	    WarnWin [_ "Please, log in first."]
	    set ret 0
	} else {
	    # may be some test like GetStatusDB to see if connection is alive ...
	}
	return $ret
    }
    proc CheckLoginAndModelID {} {
	variable _modelID

	if { ![ CheckLogin]} { return 0}
	set ret 1
	if { $_modelID == ""} {
	    WarnWin [_ "Please, open a model first."]
	    set ret 0
	} else {
	    # may be some test like GetStatusDB to see if connection is alive ...
	}
	return $ret
    }

    proc OGLdrawLists {} {
	variable _ogl
	if { [ info exists _ogl]} {
	    foreach name [ array names _ogl list,*] {
		GiD_OpenGL draw -calllist $_ogl($name)
	    }
	}
    }
    proc OGLregisterDraw {} {
	variable _ogl
	if { [ info exists _ogl]} {
	    set _ogl(register_id) [ GiD_OpenGL register ::VELaSSCo::OGLdrawLists]
	}
    }
    proc OGLunregisterAndDeleteLists {} {
	variable _ogl
	if { [ info exists _ogl]} {
	    if { [ info exists _ogl(register_id)] && ( $_ogl(register_id) != -1)} {
		GiD_OpenGL unregister $_ogl(register_id)
		set _ogl(register_id) -1
	    }
	    # now deleting lists
	    foreach name [ array names _ogl list,] {
		GiD_OpenGL draw -deletelists $_ogl($name) 1
	    }
	}
    }

    proc OGLcreateBBoxList { bbox} {
	variable _ogl
	
	# create list
	set _ogl(list,bbox) [ GiD_OpenGL draw -genlists 1]
	foreach var [ list min_x min_y min_z max_x max_y max_z] val $bbox {
	    set $var $val
	}
	
	set vertices(0) [ list $min_x $min_y $min_z]
	set vertices(1) [ list $max_x $min_y $min_z]
	set vertices(2) [ list $max_x $max_y $min_z]
	set vertices(3) [ list $min_x $max_y $min_z]
	set vertices(4) [ list $min_x $min_y $max_z]
	set vertices(5) [ list $max_x $min_y $max_z]
	set vertices(6) [ list $max_x $max_y $max_z]
	set vertices(7) [ list $min_x $max_y $max_z]
	set lst_lines [ list \
			    [ list 0 1] [ list 1 2] [ list 2 3] [ list 3 0] \
			    [ list 0 4] [ list 1 5] [ list 2 6] [ list 3 7] \
			    [ list 4 5] [ list 5 6] [ list 6 7] [ list 7 4]]

	GiD_OpenGL draw -newlist $_ogl(list,bbox) compile
	# somewhat dark magenta
	GiD_OpenGL draw -color "0.69841 0.190464 0.69841"
	GiD_OpenGL draw -begin lines
	foreach ln $lst_lines {
	    foreach nd $ln {
		GiD_OpenGL draw -vertex $vertices($nd)
	    }
	}
	# end lines
	GiD_OpenGL draw -end
	# end of  list
	GiD_OpenGL draw -endlist

	#register OpenGL proc if not already registered
	OGLregisterDraw
    }

    proc GIDcreateMeshFromObjArray { mesh_obj mesh_name element_type num_nodes_per_element} {
	variable _cache
	variable _selected
	variable _dummy_mesh_name

	# returns a list of:
	# list_of_vertices = list of id's + list of coords
	# list_of_triangle_nodes
	# example: { { { 1 2 3} { 0.0 0.0 0.0 1.0 0.0 0.0 0.0 1.0 0.0}} { 1 2 3}}
	set vertices [ lindex $mesh_obj 0]
	set elem_conectivity [ lindex $mesh_obj 1]
	set elem_properties {}
	if { [ llength $mesh_obj] > 2} {
	    # radii for the Spheres
	    set elem_properties [ lindex $mesh_obj 2]
	}
	set lst_node_ids [ lindex $vertices 0]
	set lst_coords [ lindex $vertices 1]
	set _cache(VerticesID) $lst_node_ids
	#W "GIDcreateMeshFromObjArray $mesh_name"
	#W "   setting _cache(VerticesID) with [ lrange $lst_node_ids 0 150] ..."
	# still there is one convertion UpdateStringFromLongLongArray

	# delete dummy mesh used to create result menus...
	set VersionRequired "12.1.10"
	set comp -1
	catch { 
	    set comp [ GiDVersionCmp $VersionRequired]
	}
	if { $comp == -1} {
	    # GiD_MeshPost has a bug in versions 12.1.9 and lower, so vertices ID needs a conversion:
	    set zeroed_id_list [ lmap id $lst_node_ids { expr $id - 1}]
	    set zeroed_triangle_list [ lmap id $elem_conectivity { expr $id - 1}]
	    # but using this conversion transforms the objarray into a TclString ... double the memory required
	    GiD_MeshPost create $mesh_name $element_type $num_nodes_per_element \
		-zero_based_array $zeroed_id_list $lst_coords {} $zeroed_triangle_list $elem_properties
	    set _cache(InGiD,Mesh,$mesh_name) $mesh_name
	    set _cache(InGiD,MeshElemType,$mesh_name) $element_type
	} else {
	    GiD_MeshPost create $mesh_name $element_type $num_nodes_per_element \
		$lst_node_ids $lst_coords {} $elem_conectivity $elem_properties
	    set _cache(InGiD,Mesh,$mesh_name) $mesh_name
	    set _cache(InGiD,MeshElemType,$mesh_name) $element_type
	}
	foreach res_name $_cache(ListOfResults,$_selected(analysis)) {
	    ExpandResultsInCurrentStep $_selected(analysis) $_selected(step) $res_name $lst_node_ids
	}
	# no need to delete, as it does not appear in Layer window
	# GiD_Process Mescape Utilities Delete VolumeSets $_dummy_mesh_name Yes escape escape
	# set _cache(InGiD,Mesh,$_dummy_mesh_name) 0
	# set _cache(InGiD,Mesh,$_dummy_mesh_name) ""
    }

    proc OGLcreateBoundaryMeshListFromObjArray { mesh_obj} {
	variable _ogl
	variable _modelInfo
	
	# create list
	set _ogl(list,boundaryMesh) [ GiD_OpenGL draw -genlists 1]
	set vertices [ lindex $mesh_obj 0]
	set triangles [ lindex $mesh_obj 1]
	set lst_node_ids [ lindex $vertices 0]
	set lst_coords [ lindex $vertices 1]

	# there are convertions
	#  UpdateStringFromLongLongArray and
	#  UpdateStringFromDoubleArray
	# which can be avoided using 
	#  objarray get $lst_node_ids $idx
	#  objarray get $lst_coords $idx
	#  objarray get $triangles $idx
	# or implementing a new function
	#   objarray map tri_node $lst_triangles {GiD_OpenGL draw -vertex $tri_node}
	#   line lmap

	set min_x    1e+38
	set max_x   -1e+38
	set min_y    1e+38
	set max_y   -1e+38
	set min_z    1e+38
	set max_z   -1e+38
	foreach id $lst_node_ids [ list x y z] $lst_coords {
	    set vertex_map($id) [ list $x $y $z]
	    if { $x < $min_x} { set min_x $x}
	    if { $x > $max_x} { set max_x $x}
	    if { $y < $min_y} { set min_y $y}
	    if { $y > $max_y} { set max_y $y}
	    if { $z < $min_z} { set min_z $z}
	    if { $z > $max_z} { set max_z $z}
	}
	# define the zoom_frame box
	if { ![ info exists _modelInfo] || ![ info exists _modelInfo(bbox)]} {
	    set _modelInfo(bbox) [ list $min_x $min_y $min_z $max_x $max_y $max_z]
	}

	GiD_OpenGL draw -newlist $_ogl(list,boundaryMesh) compile
	# somewhat dark green
	GiD_OpenGL draw -color "0.0 0.0 0.0"
	GiD_OpenGL draw -begin lines
	foreach [ list n1 n2 n3] $triangles {
	    GiD_OpenGL draw -vertex $vertex_map($n1)
	    GiD_OpenGL draw -vertex $vertex_map($n2)
	    GiD_OpenGL draw -vertex $vertex_map($n2)
	    GiD_OpenGL draw -vertex $vertex_map($n3)
	    GiD_OpenGL draw -vertex $vertex_map($n3)
	    GiD_OpenGL draw -vertex $vertex_map($n1)
	}

	GiD_OpenGL draw -end

	GiD_OpenGL doscrzoffset true
	GiD_OpenGL draw -color "0.8 0.8 0.6"
	GiD_OpenGL draw -begin triangles
	foreach [ list n1 n2 n3] $triangles {
	    GiD_OpenGL draw -vertex $vertex_map($n1)
	    GiD_OpenGL draw -vertex $vertex_map($n2)
	    GiD_OpenGL draw -vertex $vertex_map($n3)
	}
	GiD_OpenGL draw -end
	GiD_OpenGL doscrzoffset false

	# end of  list
	GiD_OpenGL draw -endlist

	#register OpenGL proc if not already registered
	OGLregisterDraw
    }


    # useof: can be "GEOMETRYUSE", "MESHUSE", "POSTUSE" or "GRAPHUSE". 
    proc TclCalcModelBoundaries { useof} {
	variable _modelInfo
	if { [ info exists _modelInfo] && [ info exists _modelInfo(bbox)]} {
	    return $_modelInfo(bbox)
	} else {
	    # return [ list 1e+38 1e+38 1e+38 -1e+38 -1e+38 -1e+38]
	    # return an empty list, otherwise it will use it and not the internal one!
	    return {}
	}
    }

    proc DoGetBoundingBox {} {
	variable _sessionID
	variable _modelID
	variable _modelInfo

	if { ![ CheckLoginAndModelID]} { return}
	WaitState
	Verbose 1 "calling gid_velassco GetBoundingBox $_sessionID $_modelID"
	set modelBBox [ gid_velassco GetBoundingBox $_sessionID $_modelID]
	# set modelBBox [ list 0.0 0.0 0.0 1.0 1.0 1.0]
	Verbose 1 "selected model = BBox = $modelBBox"
	
	# register a drawopengl procedure to draw the bounding box
	if { [ llength $modelBBox] == 6} {
	    OGLcreateBBoxList $modelBBox
	}
	set _modelInfo(bbox) $modelBBox
	GiD_Process 'Zoom Frame
	EndWaitState
    }

    proc GetListOfTetrahedraMeshes {} {
	variable _sessionID
	variable _modelID
	variable _cache

	if { ![ info exists _cache(GetListOfMeshes,Results)]} {
	    # Verbose 1 "getting Mesh information list"
	    # # get the list of meshes
	    # # testing GetListOfAnalyses
	    # set lst_analyses [ gid_velassco GetListOfAnalyses $_sessionID $_modelID]
	    # set analysis [ lindex $lst_analyses 0]
	    # # testing GetListOfSteps
	    # set lst_steps [ gid_velassco GetListOfSteps $_sessionID $_modelID $analysis]
	    # set step [ lindex $lst_steps 0]
	    # 
	    # # testing GetListOfMeshes
	    # if { "$analysis" != "DEM"} {
	    # 	# get information about static meshes
	    # 	set analysis ""
	    # 	set step 0.0
	    # }
	    # set lst_meshes [ gid_velassco GetListOfMeshes $_sessionID $_modelID $analysis $step]
	    # 
	    # set _cache(GetListOfMeshes,Query) "GetListOfMeshes $_sessionID $_modelID $analysis $step"
	    # set _cache(GetListOfMeshes,Results) $lst_meshes
	    GetModelInfo 0
	}

	# the information of the meshes is like this:
	#   NumberOfMeshes: 11
	#   Name: volumes
	#   ElementType: Tetrahedra
	#   NumberOfVerticesPerElement: 0
	#   NumberOfVertices: 3391
	#   NumberOfElements: 656399645689249792
	#   Units: 
	#   Color: 
	#   MeshNumber: 1
	#   CoordsName: 
	#   Name: cil
	#   ElementType: Tetrahedra
	#   NumberOfVerticesPerElement: 0
	#   NumberOfVertices: 0
	#   NumberOfElements: 960
	#   Units: 
	#   Color: 
	#   MeshNumber: 2
	#   CoordsName: 
	#   Name: wake
	#   ...
	set lst_tetra_meshes {}
	set lst_MeshInfo [ split $_cache(GetListOfMeshes,Results) "\n"]
	foreach mesh_info $lst_MeshInfo {
	    set key [ lindex $mesh_info 0]
	    if { $key == "Name:"} {
		set last_name [ lrange $mesh_info 1 end]
	    } elseif { $key == "ElementType:"} {
		set elType [ string tolower [ lindex $mesh_info 1]]
		if { $elType == "tetrahedra"} {
		    lappend lst_tetra_meshes $last_name
		}
	    }
	}
	return $lst_tetra_meshes
    }

    proc GIDdeleteMeshes {} {
	variable _cache
	variable _dummy_mesh_name
	
	foreach key [ array names _cache InGiD,Mesh,*] {
	    set mesh_name $_cache($key)
	    if { $mesh_name == $_dummy_mesh_name} {
		# do not delete this one!
		continue
	    }
	    if { $mesh_name != ""} {
		set elType [ string tolower $_cache(InGiD,MeshElemType,$mesh_name)]
		set mesh_type VolumeSets
		set surface_types [ list triangle quadrilateral line]
		if { [ lsearch $surface_types $elType] != -1} {
		    set mesh_type SurfaceSets
		}
		GiD_Process Mescape Utilities Delete $mesh_type "$mesh_name" Yes escape escape
		unset _cache($key)
		unset _cache(InGiD,MeshElemType,$mesh_name)
	    }
	}
    }

    proc GIDcreateDEMMeshes { } {
	variable _sessionID
	variable _modelID
	variable _cache
	variable _selected

	if { ![ CheckLoginAndModelID]} { return}
	if { ![ info exists _cache(GetListOfMeshes,Results)]} {
	    # Verbose 1 "getting Mesh information list"
	    # # get the list of meshes
	    # # testing GetListOfAnalyses
	    # set lst_analyses [ gid_velassco GetListOfAnalyses $_sessionID $_modelID]
	    # set analysis [ lindex $lst_analyses 0]
	    # # testing GetListOfSteps
	    # set lst_steps [ gid_velassco GetListOfSteps $_sessionID $_modelID $analysis]
	    # set step [ lindex $lst_steps 0]
	    # 
	    # # testing GetListOfMeshes
	    # if { "$analysis" != "DEM"} {
	    # 	# get information about static meshes
	    # 	set analysis ""
	    # 	set step 0.0
	    # }
	    # set lst_meshes [ gid_velassco GetListOfMeshes $_sessionID $_modelID $analysis $step]
	    # 
	    # set _cache(GetListOfMeshes,Query) "GetListOfMeshes $_sessionID $_modelID $analysis $step"
	    # set _cache(GetListOfMeshes,Results) $lst_meshes
	    GetModelInfo 0
	}

	WaitState

	set lst_mesh_names {}
	set lst_MeshInfo [ split $_cache(GetListOfMeshes,Results) "\n"]

	# repack information into an array:
	set no_name_mesh -UNKNOWN-
	set mesh_name ""
	set err_str ""
	set NumberOfMeshes 0
	foreach mesh_info_token $lst_MeshInfo {
	    if { [ llength $mesh_info_token] >= 2} {
		set key [ lindex $mesh_info_token 0]
		if { $key == "NumberOfMeshes:"} {
		    # first key: numbet of total meshes
		    set NumberOfMeshes [ lindex $mesh_info_token 1]

		} elseif { $key == "Name:"} {
		    set mesh_name [ lrange $mesh_info_token 1 end]
		    set mesh_info($mesh_name,$key) [ lrange $mesh_info_token 1 end]
		    lappend lst_mesh_names $mesh_name

		} else {
		    if { ( $mesh_name != "")} {
			set mesh_info($mesh_name,$key) [ lrange $mesh_info_token 1 end]
		    } else {
			set mesh_name $no_name_mesh
			set mesh_info($mesh_name,$key) [ lrange $mesh_info_token 1 end]
			set mesh_info($mesh_name,Name:) $mesh_name
		    }
		}
	    } else {
		append err_str "Unknown information '$mesh_info_token'\n"
	    }
	}
	if { [ info exists mesh_info($no_name_mesh,Name:)]} {
	    lappend lst_mesh_names $no_name_mesh
	}

	set gm_analysis ""
	set gm_step 0.0
	if { $_cache(InGiD,DynamicMesh)} {
	    set gm_analysis $_selected(analysis)
	    set gm_step $_selected(step)
	}

	# W "creating DEM meshes ..."
	set _cache(VerticesID) {}
	# now create the GiD_meshes
	foreach mesh_name $lst_mesh_names {
	    set elType [ string tolower $mesh_info($mesh_name,ElementType:)]
	    if { $elType == "sphere"} {
		set doing "gid_velassco GetMeshDrawData $_sessionID $_modelID $gm_analysis $gm_step $mesh_name"
		set mesh_num $mesh_info($mesh_name,MeshNumber:)
		Verbose 1 "doing gid_velassco GetMeshDrawData $_sessionID $_modelID $gm_analysis $gm_step $mesh_name"
		set lst_mesh_obj [ gid_velassco GetMeshDrawData $_sessionID $_modelID $gm_analysis $gm_step $mesh_name]
		GiD_MeshPost create $mesh_name Sphere 1 {*}$lst_mesh_obj
		set _cache(VerticesID) [ concat $_cache(VerticesID) [ lindex $lst_mesh_obj 0]]
		set _cache(InGiD,Mesh,$mesh_name) $mesh_name
		set _cache(InGiD,MeshElemType,$mesh_name) $elType
		foreach res_name $_cache(ListOfResults,$_selected(analysis)) {
		    ExpandResultsInCurrentStep $_selected(analysis) $_selected(step) $res_name $_cache(VerticesID)
		}
	    } else {
		# not implemented yet
	    }

	    # proc GIDcreateMeshFromObjArray { mesh_obj mesh_name element_type num_nodes_per_element}
	}

	EndWaitState
    }

    proc DoGetBoundaryOfAMesh {} {
	variable _sessionID
	variable _modelID
	variable _cache

	if { ![ CheckLoginAndModelID]} { return}

	WaitState
	foreach selected_mesh_name [ GetListOfTetrahedraMeshes] {
	    Verbose 1 "calling gid_velassco GetBoundaryOfAMesh $_sessionID $_modelID $selected_mesh_name"
	    set mesh_data [ gid_velassco GetBoundaryOfAMesh $_sessionID $_modelID $selected_mesh_name]
	    # set modelBBox [ list 0.0 0.0 0.0 1.0 1.0 1.0]
	    set vertex_info [ lindex $mesh_data 0]
	    set vertex_ids [ lindex $vertex_info 0]
	    set triangle_nodes [ lindex $mesh_data 1]
	    set num_vertices [ objarray length $vertex_ids]
	    set num_triangles [ expr [ objarray length $triangle_nodes] / 3]
	    Verbose 1 " got boundary of '$selected_mesh_name' with $num_vertices vertices and $num_triangles faces"
	    
	    # it should have create a GiD_PostMesh but in the meantine ...
	    # register a drawopengl procedure to draw the boundary mesh
	    if { [ llength $mesh_data] == 2} {
		# GIDcreateMesh needs a correction in GiD 12.1.10
		# set VersionRequired "12.1.10"
		# set comp -1
		# catch { 
		#     # set comp [ GiDVersionCmp $VersionRequired]
		# }
		# update
		# if { $comp == -1 } {
		#     # use tcl drawopengl fallback.
		#     OGLcreateBoundaryMeshListFromObjArray $mesh_data
		# } else {
		    GIDcreateMeshFromObjArray $mesh_data "$selected_mesh_name boundary" Triangle 3
		# }
	    }
	    GiD_Process 'Zoom Frame
	}
	EndWaitState
    }

    proc Login { { w_bar_icon ""} { w .gid.wVelasscoLogin}} {
	variable _status_wid
	variable _login_icon_wid
	if { $w_bar_icon != ""} {
	    set _login_icon_wid $w_bar_icon
	}

	if { [ winfo exists $w]} {
	    destroy $w
	}
	toplevel $w
	wm title $w [_ "VELaSSCo access window..."]
	# CreateApplyCloseDecorationWindow $w $w.top $w.bot \
	    # 	{WarnWin Accept} \
	    # 	"destroy $w"
	CreateDecorationWindow $w  $w.top $w.bot\
	    [ list [_ "Login"] [_ "Status"] [_ "Logout"]] \
	    [ list [ list VELaSSCo::DoLogin $w] \
		  VELaSSCo::DoStatusDB VELaSSCo::DoLogout] \
	    [ list <Return> ""]
	ttk::frame $w.top.header
	ttk::label $w.top.header.l1 -image [ VELaSSCo::GetImage VelasscoBig67x48.png]
	ttk::label $w.top.header.l2 -text [_ "VELaSSCo login"]
	grid $w.top.header.l1 $w.top.header.l2 -padx 4 -pady 4 -sticky ew

	ttk::frame $w.top.login
	foreach id [ list 1 2 3 4] \
	    var_name [ list "Username" "Password" "Host" "Port"] \
	    txt [ list [_ "Username"] [_ "Password"] [_ "VELaSSCo host"] [_ "Port"]] \
	    width [ list 16 16 20 8] {
		ttk::label $w.top.login.l$id -text ${txt}:
		ttk::entry $w.top.login.e$id -width $width -textvariable ::VELaSSCo($var_name)
		grid $w.top.login.l$id $w.top.login.e$id -padx 4 -pady 4 -sticky e
		grid configure  $w.top.login.e$id -sticky w
		if { $var_name == "Password"} {
		    $w.top.login.e$id configure -show *
		}
	    }
	ttk::checkbutton $w.top.login.cbTunel -text [_ "Tunnel through"]: \
	    -variable ::VELaSSCo(Tunnel,enabled)
	grid $w.top.login.cbTunel -padx 4 -pady 4 -sticky w -columnspan 2
	set lst_w_tunnel {}
	foreach id [ list 5 6 7 8] \
	    var_name [ list "Host" "LocalPort" "Username" "Password" ] \
	    txt [ list [_ "Tunnel host"] [_ "local port"] [_ "Username"] [_ "Password"] ] \
	    width [ list 20 8 16 16] {
		ttk::label $w.top.login.l$id -text ${txt}:
		ttk::entry $w.top.login.e$id -width $width -textvariable ::VELaSSCo(Tunnel,$var_name)
		grid $w.top.login.l$id $w.top.login.e$id -padx 4 -pady 4 -sticky e
		grid configure  $w.top.login.e$id -sticky w
		lappend lst_w_tunnel $w.top.login.l$id $w.top.login.e$id
		if { $var_name == "Password"} {
		    $w.top.login.e$id configure -show *
		}
	    }
	$w.top.login.cbTunel configure -command \
	    [ list VELaSSCo::EnableDisableWidgets $lst_w_tunnel ::VELaSSCo(Tunnel,enabled)]

	frame $w.top.status -height 32 -bd 0
	ttk::label $w.top.status.txt -justify center
	grid $w.top.status.txt -padx 4 -pady 4
	
	set _status_wid $w.top.status

	grid columnconfigure $w.top.login 0 -weight 1
	grid columnconfigure $w.top.login 1 -weight 1
	grid rowconfigure $w.top.login 1 -weight 1
	grid rowconfigure $w.top.login 2 -weight 1

	grid columnconfigure $w.top.status 0 -weight 1
	grid rowconfigure $w.top.status 0 -weight 1

	grid $w.top.header -sticky ew
	grid $w.top.login -sticky ew
	grid $w.top.status -sticky ew
	grid columnconfigure $w.top 0 -weight 1
	grid rowconfigure $w.top 1 -weight 1

	# default values
	# foreach var_name [ list "Username" "Password" "Host" "Port"] \
	#     value [ list miguel 1234velassco pez001 9990] {
	# 	set ::VELaSSCo($var_name) $value
	#     }
	# set ::VELaSSCo(Tunnel,enabled) 1
	# foreach var_name [ list "Host" "LocalPort" "Username" "Password" ] \
	#     value [ list acuario.cimne.upc.edu 9990 miguel 1234] {
	# 	set ::VELaSSCo(Tunnel,$var_name) $value
	#     }
	# set ::VELaSSCo(SelectedModel) ""
	if { ![ info exists ::VELaSSCo(Username)]} {
	    # value [ list andreas 1234 localhost 26267]
	    foreach var_name [ list "Username" "Password" "Host" "Port"] \
		value [ list velassco VELaSSCo localhost 26267] {
		    set ::VELaSSCo($var_name) $value
		}
	    set ::VELaSSCo(Tunnel,enabled) 0
	    foreach var_name [ list "Host" "LocalPort" "Username" "Password" ] \
		value [ list acuario.cimne.upc.edu 9990 miguel ****] {
		    set ::VELaSSCo(Tunnel,$var_name) $value
		}
	}
	set ::VELaSSCo(SelectedModel) ""

	# enable/disable tunel widgets ...
	VELaSSCo::EnableDisableWidgets $lst_w_tunnel ::VELaSSCo(Tunnel,enabled)
	ShowLoginStatus
    }

    proc GetResultForVertices { result_name} {
	variable _sessionID
	variable _modelID
	variable _cache 
	variable _selected
	variable _busy
	set res_values [ list {} {}]
	if { ![ CheckLoginAndModelID]} { return $res_values}

	# get result values from VELaSSCo and
	# updates them into GiD:
	#  GetResultFromVerticesID sessionID modelID resultID analysisID lst_vertexIDS timeStep

	set sel_analysis $_selected(analysis)
	set sel_step $_selected(step)
	set res_type $_cache($sel_analysis,$sel_step,$result_name,Type)
	set res_num_comp $_cache($sel_analysis,$sel_step,$result_name,NumberOfComponents)

	Verbose 1 "GetResultFromVerticesID $_sessionID $_modelID \"$result_name\" $res_type $res_num_comp \"$_selected(analysis)\" N_vertices $_selected(step)"
	if { [ info exists _cache(VerticesID)] && ( [ llength $_cache(VerticesID)] != 0)} {
	    Verbose 1 "    retrieving results for [ lrange $_cache(VerticesID) 0 30] ..."
	    set res_values [ gid_velassco GetResultFromVerticesID $_sessionID $_modelID \
				 "$result_name" $res_type $res_num_comp \
				 "$_selected(analysis)" $_cache(VerticesID) $_selected(step)]
	    # returns 2 lists: lst_ids and lst_results
	    # W "lst_ids size = [ llength [ lindex $res_values 0]]"
	    # W "lst_res size = [ llength [ lindex $res_values 1]]"
	} else {
	    if { ![ info exists _cache(VerticesID)]} {
		W " no _cache(VerticesID)"
	    } else {
		W "_cache(VerticesID) is empty? = -$_cache(VerticesID)-"
	    }
	}
	return $res_values
    }

    proc GetResultForNodeEvolution { result_name lst_vertex_ids} {
	variable _sessionID
	variable _modelID
	variable _cache 
	variable _selected
	variable _busy
	variable _preferences
	set res_values [ list {} {}]
	if { ![ CheckLoginAndModelID]} { return $res_values}

	# get result values from VELaSSCo and
	# updates them into GiD:
	#  GetResultFromVerticesID sessionID modelID resultID analysisID lst_vertexIDS timeStep

	set sel_analysis $_selected(analysis)
	set sel_step $_selected(step)
	set res_type $_cache($sel_analysis,$sel_step,$result_name,Type)
	set res_num_comp $_cache($sel_analysis,$sel_step,$result_name,NumberOfComponents)

	set res_values {}
	if { [ info exists _cache(ListOfSteps)] && ( [ llength $_cache(ListOfSteps)] >= 1) } {
	    # the vquery does not accept a list of steps as a parameter ...

	    if { [ info exists _preferences(NumberOfStepsForNode)]} {
		set steps_for_node_evolution $_preferences(NumberOfStepsForNode)
	    } else {
		# set steps_for_node_evolution all
		set steps_for_node_evolution 5
		set _preferences(NumberOfStepsForNode) $steps_for_node_evolution
	    }
	    # for the use-case as it's too slow: do it only for 5 steps from _selected(step)
	    set step_list $_cache(ListOfSteps)
	    if { ( [ string tolower $steps_for_node_evolution] != "all") && ( $steps_for_node_evolution > 0)} {
		if { [ llength $_cache(ListOfSteps)] > $steps_for_node_evolution} {
		    #select 5 steps
		    set idx_step [ lsearch $_cache(ListOfSteps) $sel_step]
		    set len_lst_step [ llength $_cache(ListOfSteps)]
		    # if idx_step is among the last 5, use the last 5
		    set last_5 [ expr $len_lst_step - $steps_for_node_evolution]
		    if { $last_5 < 0} {
			set last_5 0
		    }
		    if { $idx_step > $last_5} {
			set idx_step $last_5
		    }
		    # lrange includes both ini_idx and end_idx, so using 4 instead of 5
		    set end_idx_step [ expr $idx_step + $steps_for_node_evolution - 1]
		    if { $end_idx_step > $len_lst_step} {
			set end_idx_step end
		    } 
		    set step_list [ lrange $_cache(ListOfSteps) $idx_step $end_idx_step]
		}
	    }

	    # W "lst_vertex_ids = 	    $lst_vertex_ids"
	    # W "step_list = $step_list"

	    foreach step $step_list {
		Verbose 1 "GetResultFromVerticesID $_sessionID $_modelID \"$result_name\" $res_type $res_num_comp \"$_selected(analysis)\" $lst_vertex_ids $step"
		set result_ids_values [ gid_velassco GetResultFromVerticesID $_sessionID $_modelID \
					    "$result_name" $res_type $res_num_comp \
					    "$_selected(analysis)" $lst_vertex_ids $step]
		set graf_ids [ lindex $result_ids_values 0]
		set graf_values [ lindex $result_ids_values 1]
		# although it's only one value it's a nested list: { {0.0 0.0 0.0}}
		lappend res_values [ lindex $graf_values 0]
	    }
	} else {
	    WarnWin [_ "Please, get the model information first."]
	}
	# returns 2 lists: lst_ids and lst_results
	# W "lst_ids size = [ llength [ lindex $res_values 0]]"
	# W "lst_res size = [ llength [ lindex $res_values 1]]"
	return [ list $step_list $res_values]
    }

    proc doNodeEvolution { action res_name comp_name name1 name2 op} {
	variable _cache
	variable _selected
	trace remove variable ::GidPriv(selection) write [ list ::VELaSSCo::catchDoNodeEvolution $action $res_name $comp_name]
	# set res_name [ lindex $_cache(ListOfResults,$_selected(analysis)) 0]
	set sel_analysis $_selected(analysis)
	set sel_step $_selected(step)
	set idx_comp 0
	if { $comp_name != ""} {
	    set idx_comp [ lsearch $_cache($sel_analysis,$sel_step,$res_name,ComponentNames) $comp_name]
	    if { $idx_comp == -1} {
		set idx_comp 0
	    }
	}

	# $::GidPriv(selection) is of the form "Nodes 1" (with the quotes)
	set node_selection [ regsub -all \" $::GidPriv(selection) { }]
	set node_id [ lindex $node_selection 1]
	set graf_values [ GetResultForNodeEvolution $res_name $node_id]
	set step_values [ lindex $graf_values 0]
	set res_values [ lindex $graf_values 1]
	# W "graf_values = $graf_values"

	# create a graph file:
	set tmp_filename [ file join [ GiD_Info project TmpDirectory] "Evolution_${res_name}_node_${node_id}.grf"]
	set fo [ open $tmp_filename w]
	puts $fo "# Graph: \"Evolution of $res_name on node $node_id\""
	puts $fo "# X: \"Step value\" Y: \"$res_name\""
	foreach step $step_values value $res_values {
	    puts $fo "$step [ lindex $value $idx_comp]"
	}
	close $fo
	# now read the graph in GiD
	GiD_Process Mescape Results Graphs ReadGraph "$tmp_filename" escape
	# W "file $tmp_filename created"
	# processing GiD_Process takes some time and severl Tcl event loops ...
	after 3000 [ list file delete -force $tmp_filename]
    }

    proc catchDoNodeEvolution { action res_name comp_name name1 name2 op} {
	set err [
		 catch {
		     doNodeEvolution $action $res_name $comp_name $name1 $name2 $op
		 } err_txt
		 ]
	if { $err} {
	    W "doNodeEvolution $action $res_name $comp_name $name1 $name2 $op = \n\t$err_txt"
	}
    }

    proc startNodeEvolution { result_name { component_name ""}} {
	variable _cache
	variable _preferences
	set parent_w .gid
	if { ![ info exists _cache(ListOfSteps)] || ( [ llength $_cache(ListOfSteps)] < 1) } {
	    WarnWin [_ "Please, get the model information first."]
	    return
	}

	set res_name [ fromGiDEncoding $result_name] 
	set comp_name [ fromGiDEncoding $component_name]

	set w $parent_w.nodeEntry
	#[ VELaSSCo::GetImage VelasscoBig67x48.png]
	if { [ info exists _preferences(NumberOfStepsForNode)]} {
	    set number_steps $_preferences(NumberOfStepsForNode)
	} else {
	    # set steps_for_node_evolution all
	    set number_steps 5
	    set _preferences(NumberOfStepsForNode) $steps_for_node_evolution
	}

	if { $component_name != ""} {
	    set dialog_txt [_ "Enter the Node ID / Vertex ID \nto do a \n'$res_name' ('$comp_name')\n evolution along $number_steps steps:"]
	} else {
	    set dialog_txt [_ "Enter the Node ID / Vertex ID \nto do a \n'$res_name'\n evolution along $number_steps steps:"]
	}
	set vertexID [ tk_dialogEntryRAM $w [_ "Vertex ID"] \
			   $dialog_txt \
			   question int+ 1
		       ]
	if { ( $vertexID == "") || ( $vertexID == "--CANCEL--")} {
	    return
	}

	# if node is in gid:
	# GiD_Process 'SelectEntities JustOne Nodes
	# now:
	trace add variable ::GidPriv(selection) write [ list ::VELaSSCo::catchDoNodeEvolution NodeEvolution \
							    $res_name $comp_name]
	# $::GidPriv(selection) is of the form "Nodes 1" (with the quotes)
	set ::GidPriv(selection) "Nodes $vertexID"
    }


    proc AddGraphsResultsToMenu { w basecommand que_res } {
	global VelasscoViewMenuPriv
	
	# registramos los menus
	# if { [ esMac]} {
	#         registerPostGraphsResultsMenu $w $basecommand $que_res
	#         $w configure -postcommand ""
	# }
	
	set all_anal [lsort -dictionary [ GiD_Info postprocess get all_analysis]]
	# set all_anal [ GiD_Info postprocess get all_analysis]
	if { $all_anal == ""} {
	    set msg [_ "Please load a MODEL first"]
	} else {
	    set msg [_ "Please select a STEP first"]
	}
	set cur_analysis [ GiD_Info postprocess get cur_analysis]
	if { $cur_analysis == "" } {
	    $w delete 0 end
	    $w add command -label $msg -state disabled
	    return
	}
	
	set cur_step [ GiD_Info postprocess get cur_step $cur_analysis]
	if { $cur_step == ""} {
	    $w delete 0 end
	    $w add command -label [_ "Please select a STEP first"] -state disabled
	    return
	}
	
	set all_res [ lsort -dictionary [ GiD_Info postprocess get cur_results_list $que_res]]
	if { $all_res == ""} {
	    $w delete 0 end
	    $w add command -label [_ "No useful results found"] -state disabled
	    return
	}
	
	if { $que_res == "Integrate_Vector_Normal" || $que_res == "Integrate_Vector_Tangential" } {
	    set also_components 0
	} else {
	    set also_components 1
	}
	
	## miramos si el menu que ya tenemos aun es valido
	# if { ![esMac]}  # en mac solo nos llaman cuando hay que rehacer los menus
	if { 1} { # en mac solo nos llaman cuando hay que rehacer los menus
	    # para el conyazo de windows, que es mas lento que una tortuga con reuma
	    if { ( $cur_analysis != "") && ( $cur_step != "") && ( $all_res != "") } {
		if { [ info exists VelasscoViewMenuPriv($w,$que_res//analstep)] && \
			 [ info exists VelasscoViewMenuPriv($w,$que_res//allres)] } {
		    if { ![ string compare $VelasscoViewMenuPriv($w,$que_res//analstep) "$cur_analysis $cur_step" ] || \
			     ![ string compare $VelasscoViewMenuPriv($w,$que_res//allres) $all_res] } {
			# ahora a mirar los componentes
			set igual 1
			foreach i $all_res {
			    if { $also_components } {
				set all_comp [ GiD_Info postprocess get cur_components_list $i ]
			    } else {
				set all_comp ""
			    }
			    regsub -all { } $i {_} idx
			    if { [ info exists VelasscoViewMenuPriv($w,$que_res//$idx)] } {
				if { [ string compare $VelasscoViewMenuPriv($w,$que_res//$idx) $all_comp]} {
				    # no son iguales
				    set igual 0
				    break
				}
			    } else {
				# no son iguales
				set igual 0
				break
			    }
			}
			if { $igual} {
			    return
			}
		    }
		}
	    }
	}
	
	# para ahorrarnos construir cosas
	set VelasscoViewMenuPriv($w,$que_res//analstep) "$cur_analysis $cur_step"
	set VelasscoViewMenuPriv($w,$que_res//allres) $all_res
	foreach i $all_res {
	    if { $also_components } {
		set all_comp [ GiD_Info postprocess get cur_components_list $i ]
	    } else {
		set all_comp ""
	    }
	    regsub -all { } $i {_} idx
	    set VelasscoViewMenuPriv($w,$que_res//$idx) $all_comp
	}
	
	$w delete 0 end
	
	if { $que_res == "Line_Graph"} {
	    set eje X
	    # construyamos el menu del eje X
	    if { [ winfo exists $w.m$eje ]} {
		$w.m$eje delete 0 end
	    } else {
		menu $w.m$eje
	    }
	    set cmds {X_Variation Y_Variation Z_Variation LineVariation}
	    set labels [ list [_ "X Variation"] [_ "Y Variation"] [_ "Z Variation"] [_ "Line Variation"]]
	    set values [ list "X_Variation" "Y_Variation" "Z_Variation" "LineVariation"]
	    set UnderLineList {}
	    if { ![ info exists ::VelasscoViewMenuPriv(LineGraphXAxis) ]} {
		set ::VelasscoViewMenuPriv(LineGraphXAxis) "LineVariation"
	    }
	    foreach i $cmds l $labels v $values {
		set under [FindUnderChar l UnderLineList]
		$w.m$eje add checkbutton -label $l -underline $under \
		    -onvalue $v -variable ::VelasscoViewMenuPriv(LineGraphXAxis)
	    }
	    $w add cascade -label [_ "Set X axis"] -menu $w.m$eje
	}
	
	set entrada_ventana 0
	if { $all_res != ""} {
	    foreach i $all_res {
		if { $also_components } {
		    if { $que_res != "Point_Complex_Evolution"} {
			set all_comp [ GiD_Info postprocess get cur_components_list $i ]
		    } else {
			set all_comp [ GiD_Info postprocess get cur_complex_components_list $i ]
		    }
		} else {
		    set all_comp ""
		}
		
		set wm [ GetResultMenuFromPath $w $i ""]
		if { $wm == ""} {
		    set entrada_ventana 1
		    break
		}
		
		set txt $i
		if { $wm != $w} {
		    if { [regexp {^/?/?(.*)//(.*?)$} $i dum path txt_res] } {
			set txt $txt_res
		    }                 
		}
		set txt_res [TranslateResultName $txt]
		
		set create_submenu 0
		if { [ llength $all_comp] > 1 } {
		    set create_submenu 1
		}
		
		# check if althought it's a scalar there are some created components elsewhere
		if { [ llength $all_comp] == 1} {
		    set root_res ${i}//
		    if { [ string first $root_res $all_res] != -1} {
			# may be of the form t//something
			# if { [ lsearch $all_res $root_res] != -1} 
			set create_submenu 1
			set txt_res [TranslateResultName $txt]
		    }
		    # WarnWinText "looking for '$root_res' in '$all_res' = [ string first $root_res $all_res]"
		}
		
		if { $create_submenu } {
		    # add component's menu
		    regsub -all {\.} $i {_} menu_name
		    $wm add cascade -label $txt_res -menu $wm.m$menu_name
		    if { [ winfo exists $wm.m$menu_name ]} {
			$wm.m$menu_name delete 0 end
		    } else {
			menu $wm.m$menu_name
		    }
		    foreach j $all_comp {
			set txt_comp [TranslateResultName $j]                   
			set do_graph_cmd [list {*}$basecommand $i $j]                    
			if { $que_res == "Line_Graph"} {
			    # set do_graph_cmd [list DoLineGraph $w {*}$basecommand $i $j]
			    set do_graph_cmd [list DoLineGraph $w $basecommand $i $j]
			}
			$wm.m$menu_name add command -label $txt_comp -command $do_graph_cmd
		    }
		} else {
		    # add the command
		    set do_graph_cmd [list {*}$basecommand $i]
		    if { $que_res == "Line_Graph"} {
			# set do_graph_cmd [list DoLineGraph $w {*}$basecommand $i]
			set do_graph_cmd [list DoLineGraph $w $basecommand $i]
		    }
		    $wm add command -label $txt_res -command $do_graph_cmd
		}
	    }
	}
	
	if { $entrada_ventana} {
	    $w add separator
	    $w add command -label ...[_ "More"]... -command PostGraphs::Create
	}
    }


    proc DoNodeInfo { result_name { component_name ""}} {
	variable _sessionID
	variable _modelID
	variable _cache 
	variable _selected
	variable _busy

	set parent_w .gid

	if { ![ CheckLoginAndModelID]} { return}

	# W "DoNodeInfo:"
	# W "parent_w = $parent_w"
	# W "result_name = $result_name"
	# W "component_name = $component_name"

	# [ VELaSSCo::GetImage VelasscoBig67x48.png]
	set w $parent_w.nodeEntry
	set vertexID [ tk_dialogEntryRAM $w [_ "Vertex ID"] \
			   [_ "Enter the Node ID / Vertex ID \nto retrieve his current result values:"] \
			   question int+ 1
		       ]
	if { ( $vertexID == "") || ( $vertexID == "--CANCEL--")} {
	    return
	}

	# get result values from VELaSSCo and
	# updates them into GiD:
	#  GetResultFromVerticesID sessionID modelID resultID analysisID lst_vertexIDS timeStep

	set sel_analysis $_selected(analysis)
	set sel_step $_selected(step)
	
	set txt_output {}
	lappend txt_output "Result values for vertex ID = $vertexID"
	lappend txt_output "   for analysis = '$sel_analysis'"
	lappend txt_output "    and step = '$sel_step':"

	# foreach res_name $_cache(ListOfResults,$_selected(analysis)) {
	set res_name [ fromGiDEncoding $result_name]
	    if { ![ info exist _cache($sel_analysis,$sel_step,$res_name,Header)]} {
		# may be is a result in gauss points which are not handled at the moment
		return
	    }

	    set res_type $_cache($sel_analysis,$sel_step,$res_name,Type)
	    set res_num_comp $_cache($sel_analysis,$sel_step,$res_name,NumberOfComponents)
	    Verbose 1 "GetResultFromVerticesID $_sessionID $_modelID \"$res_name\" $res_type $res_num_comp \"$sel_analysis\" $vertexID $sel_step"
	    set res_values [ gid_velassco GetResultFromVerticesID $_sessionID $_modelID \
				 "$res_name" $res_type $res_num_comp \
				 "$_selected(analysis)" $vertexID $_selected(step)]
	    set lst_ids [ lindex $res_values 0]
	    set lst_res [ lindex $res_values 1]
	    # lst_res is the list of list of values for each vertex id, 
	    # if it's a vector: lst_res = { { vx vy vz} {vx vy vz} ... }
	    # as we're asking now for one vertex, lst_res = {{vx vy vz}}
	    # so:
	    set lst_res [ lindex $lst_res 0]
	    if { [ llength $lst_res] > 1} {
		lappend txt_output "   $res_name = "
		if { [ info exists _cache($sel_analysis,$sel_step,$res_name,ComponentNames)]} {
		    foreach val $lst_res comp $_cache($sel_analysis,$sel_step,$res_name,ComponentNames) {
			lappend txt_output "      $comp = $val "
		    }
		} else {
		    foreach val $lst_res {
			lappend txt_output "          $val "
		    }
		}
	    } else {
		lappend txt_output "   $res_name = $lst_res"
	    }
        #}
	set txt [ join $txt_output \n]
	W $txt
	InfoWin $txt
    }


    proc ShowHideLogWindow {} {
	variable _DEBUG_MODE
	if { !$_DEBUG_MODE} {
	    return
	}
	# PWViewOutputWin { filename name date { basename "" } { actualizar 1} }
	if { ![info exists ::RunProcInfo] || ( $::RunProcInfo == "")} {
	    set gid_output [ file join [ GiD_Info project TmpDirectory] tmp-gidOutput]
	    set gid_error [ file join [ GiD_Info project TmpDirectory] tmp-gidError]
	    set time [ clock seconds]
	    # trick to continue the output actualization if file is blocked or eof is reached
	
	    # drives only for windows------------------------------------------|--------------|
	    # RunProcInfo: $basename $time $uid $outputfiles $errorfiles $warningfiles
	    #{$local_drives} {$net_drives}
	    # RunProcInfo (remote): $basename $time remote $outputfiles $errorfiles $warningfiles $server \
		#                       $user $channel $state $dir $PT $args $timeaccess
	    # RunProcInfo uid name ... channel state
	    set ::RunProcInfo [ list [ list gid_velassco $time remote $gid_output $gid_error "" "Server" "User" "" RESULTSINSERVER "dir" "PT" "args" $time]]
	    PWViewOutputWin $gid_error [_ "GiD-VELaSSCo plug-in"] [ clock format $time] "" 1
	} else {
	    set ::RunProcInfo ""
	    PWViewOutput_CloseAll
	}
    }

    proc fromGiDEncoding { name} {
	regsub -all {_} $name { } ret_name
	return $ret_name
    }

    proc toGiDEncoding { name} {
	regsub -all { } $name {_} ret_name
	return $ret_name
    }

    proc isBusy { who} {
	variable _busy
	set busy 0
	if { [ info exists _busy($who)] && $_busy($who)} {
	    set busy 1
	}
	return $busy
    }
    proc setBusy { who value} {
	variable _busy
	set _busy($who) $value
    }

    proc addPendingEvent { type args} {
	variable _pending
	# W "adding pending event = $type $args"
	lappend _pending [ list $type $args]
    }

    # proc clearPendingEvents {} {
    # 	variable _pending
    # 	set _pending {}
    # }

    proc getPendingEvents {} {
	variable _pending
	set ret $_pending
	set _pending {}
	return $ret
    }

    proc managePostprocessUpdateEvent { type args} {
	variable _sessionID
	variable _modelID
	variable _cache 
	variable _selected
	variable _pending

	# do my update:
	# get the results and do a contour fill
	# W "type = $type"
	# W "args = $args"

	# check if we have something or we have something and sent it to GiD:
	if { ( $_sessionID == "") || ( $_modelID == "")} {
	    # no need to check if there is something in gid, if we don't have connection or an openmodel, then nothing...
	    # set something_in_gid 0
	    # if { [ llength [ array names _cache InGiD,*]] != 0} {
	    # 	set something_in_gid 0
	    # }
	    return
	}

	switch $type {
	    "ResultViewEvent" {
		# W "   processing ResultViewEvent"
		set res_view [ regsub -all {_} [ lindex $args 0] {}]

		# Some problems with the LoadSelectionEvent, not always is being generated by GiD, so
		set cur_analysis [ GiD_Info postprocess get cur_analysis]
		set cur_step [ GiD_Info postprocess get cur_step $cur_analysis]
		set gid_an [ fromGiDEncoding $cur_analysis]
		set _selected(analysis) $_cache($gid_an,OriginalName)
		set _selected(step)  [ fromGiDEncoding $cur_step]

		# eventually it can be done for any result visualization ...
		# W "res_view = $res_view,   llength args = [ llength $args] = $args"
		# W "ResultViewEvent, busy = [ isBusy myPostprocessUpdate]"
		# W "      isBusy CreatingDEMMeshes = [ isBusy CreatingDEMMeshes]"
		if { ![ isBusy CreatingDEMMeshes]} {
		    if { ( $res_view != "NoResult") && ( [ llength $args] >= 3)} {
			set gid_res_name [ lindex $args 1]
			set gid_comp_name [ lindex $args 2]
			set gid_res [ ::VELaSSCo::fromGiDEncoding $gid_res_name]
			set res_name $_cache($gid_res,OriginalName)
			set comp_name [ ::VELaSSCo::fromGiDEncoding $gid_comp_name]
			
			# is result cached ?
			# if not retrieve the results from VELaSSCo
			# pass them to gid 
			# and redo the contour fill
			# W "     doing AlreadyInGiD $_selected(analysis) $_selected(step) '$res_name'"
			set in_gid [ AlreadyInGiD $_selected(analysis) $_selected(step) "$res_name"]
			# W "           = $in_gid"
			if { !$in_gid} {
			    set result_ids_values [ ::VELaSSCo::GetResultForVertices $res_name]
			    # W "GetResultForVertices = [ lrange $result_ids_values 0 150] ..."
			    set lst_ids [ lindex $result_ids_values 0]
			    set lst_values [ lindex $result_ids_values 1]
			    if { ( [ llength $lst_ids] != 0) && ( [ llength $lst_values] != 0) } {
				# W "before CreateResultInGiD $_selected(analysis) $_selected(step) $res_name $lst_ids $lst_values"
				CreateResultInGiD $_selected(analysis) $_selected(step) "$res_name" $lst_ids $lst_values
				
				#redo the contour fill after finishing the event
				# GiD_Process Mescape Results ContourFill $gid_res_name $gid_comp_name

				# avoid calling this process again:
				::unregister_myPostprocessUpdate
				GiD_Process Mescape Results $res_view $gid_res_name $gid_comp_name
				::register_myPostprocessUpdate
			    } else {
				WarnWin "Error retrieving the $res_name values from VELaSSCo: $result_ids_values"
			    }
			}
		    }
		} else { 
		    # W "        --> pending [ list $type $args]"
		    addPendingEvent $type {*}$args
		}
	    }
	    "LoadSelectionEvent" {
		# W "   processing LoadSelectionEvent"
		set gid_an [ fromGiDEncoding [ lindex $args 0]]
		set _selected(analysis) $_cache($gid_an,OriginalName)
		set _selected(step)  [ fromGiDEncoding [ lindex $args 1]]
		# W "now analysis = $_selected(analysis)"
		# W "now step     = $_selected(step)"
		# in theory we must recreate the results tree
		# CreateResultTree $_selected(analysis) $lst_steps $lst_results
		# delete previous meshes

		# delete & create meshes only if we have dynamic (DEM) meshes:
		if { [ info exists _selected(analysis)] && [ info exists _cache(InGiD,DynamicMesh)] && \
			 ( ( $_selected(analysis) == "DEM") && $_cache(InGiD,DynamicMesh))} {
		    # both generate update events that causes Result_View events when the mesh is still not loaded
		    setBusy CreatingDEMMeshes 1
		    # W "   GIDdeleteMeshes ; GIDdeleteMeshes"
		    GIDdeleteMeshes
		    GIDcreateDEMMeshes
		    setBusy CreatingDEMMeshes 0
		    # W "Pending events ? $_pending"
		    # foreach ev $_pending {
		    #     set ev_type [ lindex $ev 0]
		    #     set ev_args [ lindex $ev 1]
		    #     W "managePostprocessUpdateEvent $ev_type {*}$ev_args"
		    #     managePostprocessUpdateEvent $ev_type {*}$ev_args
		    # }
		    # set _pending {}
		}
	    }
	}

	if { [ llength [ info commands ::hidden_GiDPostprocessUpdate]] != 0} {
	    ::hidden_GiDPostprocessUpdate $type $args
	}
    }

    proc modifyGiDmenus { } {
	set idx [ GiDMenu::GetOptionIndex Files [ list "Recent post files"] POST]
	if { $idx == -1 } {
	    set idx end
	}
	# set icon [ ::VELaSSCo::GetImage Velassco16.png]
	set icon [ ::VELaSSCo::GetFullImageName Velassco16.png]
	# skip the --- separator just after Recent
	incr idx
	GiDMenu::InsertOption Files [ list "Connect to VELaSSCo ..."] $idx POST ::VELaSSCo::Login \
	    "" $icon "insertafter" _
	incr idx
	GiDMenu::InsertOption Files [ list "Open VELaSSCo model ..."] $idx POST ::VELaSSCo::SelectModel \
	    "" $icon "insertafter" _
	incr idx
	GiDMenu::InsertOption Files [ list ---] $idx "POST" "" "" "" "insertafter"
    }

    proc velasscoIconBarProc { { what DEFAULT} { img_dir ""}} {
	variable _dir
	variable _icon_do_connection
	variable _icon_connected
	variable _DEBUG_MODE
	global velasscoIconBarBitmapsNames velasscoIconBarBitmapsCommands velasscoIconBarBitmapsHelp
	set debug_images {}
	set debug_commands {}
	set debug_help {}
	if { $_DEBUG_MODE} {
	    set debug_images [ list logviewer.png]
	    set debug_commands [ list {-np- ::VELaSSCo::ShowHideLogWindow}]
	    set debug_help [ list "Show/Hide the AccessLib log"]
	}
	set velasscoIconBarBitmapsNames(0) [ list Velassco24.png --- \
						 $_icon_do_connection --- \
						 open2.gif info.png \
						 volume.png boundary2d.png \
						 InfoVertex.png \
						 PostBarGraphPointEvolution.png PostBarGraphClear.png \
						 d2c.png \
						 --- \
						 settings.png \
						 {*}$debug_images \
						 status.png VelasscoStop.png]
	
	set velasscoIconBarBitmapsCommands(0) [ list {-np- ::VELaSSCo::About} "" \
						    {-np- ::VELaSSCo::Login %W} "" \
						    {-np- ::VELaSSCo::SelectModel} {-np- ::VELaSSCo::GetModelInfo 1} \
						    {-np- ::VELaSSCo::DoGetBoundingBox} {-np- ::VELaSSCo::DoGetBoundaryOfAMesh} \
						    {-npm- ::VELaSSCo::AddGraphsResultsToMenu %W \
							 ::VELaSSCo::DoNodeInfo Point_Evolution} \
						    {-npm- ::VELaSSCo::AddGraphsResultsToMenu %W \
							 ::VELaSSCo::startNodeEvolution Point_Evolution} \
						    {Results Graphs OptionsGraph ClearGraphs} \
						    {-np- ::VELaSSCo::D2C_window} \
						    "" \
						    {-np- ::VELaSSCo::PreferencesWindow} \
						    {*}$debug_commands \
						    {-np- ::VELaSSCo::DoStatusDB} {-np- ::VELaSSCo::DoStopVELaSSCo} ]
	set velasscoIconBarBitmapsHelp(0) [ list "About VELaSSCo" "" \
						"Connect to VELaSSco server" "" \
						"Open a Model" "Show Model information" \
						"GetBoundingBox" "GetBoundaryOfAMesh" \
						"Information about a node" \
						"Node Evolution (along several steps from current selected step)" \
						"Clear graphs" \
						"Discrete 2 Continuum transformation window" \
						"" \
						"set VELaSSCo preferences: connection, graph" \
						{*}$debug_help \
						"VELaSSCo status" "Stops QueryManager and StorageModule servers" ]

	# to create an empty pull-down menu for the node info
	set velasscoIconBarBitmapsNames(0,8)	""
	set velasscoIconBarBitmapsCommands(0,8) ""
	# to create an empty pull-down menu for the node evolution graphs
	set velasscoIconBarBitmapsNames(0,9)	""
	set velasscoIconBarBitmapsCommands(0,9) ""

	CreateOtherBitmaps velasscoIconBar "VELaSSCo icon bar" velasscoIconBarBitmapsNames velasscoIconBarBitmapsCommands \
	    velasscoIconBarBitmapsHelp [ file join $_dir images] velasscoIconBarBitmaps $what Post
	
	AddNewToolbar velasscoIcon_bar PostVelasscoIconBarWindowGeom ::VELaSSCo::velasscoIconBarProc "VELaSSCo icon bar"
    }

}

proc register_myPostprocessUpdate {} {
    if { [ llength [ info commands ::hidden_GiDPostprocessUpdate]] != 0} {
	# already registered, nothing to do
	return
    }
    rename ::GiDPostprocessUpdate ::hidden_GiDPostprocessUpdate
    rename ::myPostprocessUpdate ::GiDPostprocessUpdate
}
proc unregister_myPostprocessUpdate {} {
    if { [ llength [ info commands ::myPostprocessUpdate]] != 0} {
	# already unregistered, nothing to do
	return
    }
    rename ::GiDPostprocessUpdate ::myPostprocessUpdate 
    rename ::hidden_GiDPostprocessUpdate ::GiDPostprocessUpdate 
}

proc InitGIDProject { dir} {
    ::VELaSSCo::Ini $dir

    # modify menus
    ::VELaSSCo::modifyGiDmenus

    # change to postprocess
    GiD_Process Mescape Postprocess

    # Create a new toolbar
    ::VELaSSCo::velasscoIconBarProc INSIDELEFT
    after 1000 ::VELaSSCo::velasscoIconBarProc INSIDELEFT
    after 1000 register_myPostprocessUpdate
}

proc EndGIDProject { } {
    unregister_myPostprocessUpdate
    ::VELaSSCo::End
}
proc EndGIDPostProcess { } {
    unregister_myPostprocessUpdate
    ::VELaSSCo::End
}
proc TclCalcModelBoundaries { useof} {
    # useof: can be "GEOMETRYUSE", "MESHUSE", "POSTUSE" or "GRAPHUSE". 
    # check if namespace is loaded
    if { [ llength [ info commands ::VELaSSCo::TclCalcModelBoundaries]] != 0} {
	::VELaSSCo::TclCalcModelBoundaries $useof
    }
}

proc doPostprocessUpdate { type args} {
    ::VELaSSCo::managePostprocessUpdateEvent $type {*}$args
    foreach ev [ ::VELaSSCo::getPendingEvents] {
	set ev_type [ lindex $ev 0]
	set ev_args [ lindex $ev 1]
	::VELaSSCo::managePostprocessUpdateEvent $ev_type {*}$ev_args
    }
    ::VELaSSCo::setBusy myPostprocessUpdate 0
}

proc myPostprocessUpdate { type args} {
    #program the action
    if { [ ::VELaSSCo::isBusy myPostprocessUpdate]} {
	::VELaSSCo::addPendingEvent $type {*}$args
    	return
    }
    ::VELaSSCo::setBusy myPostprocessUpdate 1
    after 500 [ list doPostprocessUpdate $type {*}$args]
}
